import sys
import re
import gc
import time
from iterable_queue import IterableQueue
from multiprocessing import Process
from subprocess import check_output
from token_map import UNK
from unigram_dictionary import UnigramDictionary
import numpy as np

import os
# Only import theano and lasagne if environment permits it
exclude_theano_set = 'EXCLUDE_THEANO' in os.environ
if exclude_theano_set and int(os.environ['EXCLUDE_THEANO']) == 1:
	# Don't import theano and lasagne
	pass
else:
	# Do import theano and lasagne
	from theano import shared, function, tensor as T


class TheanoMinibatcher(object):
	'''
	This generates a theano shared variable storing the full dataset
	-- all training examples.  When the theano device setting is the
	GPU, shared variables are stored on the GPU, so this has the
	effect of loading the full dataset onto the GPU.

	One of the return values is a (set of) symbolic theano variable(s)
	corresponding to a single minibatch of the data.  This symbolic
	variable can be used to set up the training function.  What will
	happen during training is that this variable acts as a sliding
	"window" on the full dataset, selecting each minibatch in turn,
	even though the entire dataset is loaded into GPU memory.

	The indexing that causes the symbolic minibatch to address different
	parts of the dataset is itself a shared variable, and it can be
	updated using an update tuple provided to the updates list of a
	theanod function.  The necessary update tuple is also provided as
	a return value, so that it can be incorporated into the training
	function
	'''

	def __init__(self, batch_size=1000, dtype="float32", num_dims=2):
		self.batch_size = batch_size
		self.dtype = dtype
		self.num_dims = num_dims

		self._setup_batching()


	def _initialize_data_container(self, num_dims, dtype):

		# Validate num_dims
		if(num_dims < 1 or not isinstance(num_dims, int)):
			raise ValueError(
				'TheanoMinibatcher: num_dims must be an integer equal to or '
				'greater than 1.'
			)

		# Create the first dim, which houses the dataset
		data_container = []
		num_dims -= 1

		# Repeatedly add a nested dimension, so that we have num_dims nestings
		nested_container_handle = data_container
		for dim_num in range(num_dims):
			new_inner_container = []
			nested_container_handle.append(new_inner_container)
			nested_container_handle = new_inner_container

		return np.array(data_container, dtype=dtype)

	def reset(self):
		'''
		Reset the internal batch_num pointer to the start of the dataset
		'''
		self.batch_num.set_value(np.int32(0))


	def _setup_batching(self):

		# Make an empty shared variable that will store the dataset
		# Although empty, we can setup the relationship between the
		# minibatch variable and the full dataset
		self.dataset = shared(
			self._initialize_data_container(self.num_dims, self.dtype)
		)

		# Make minibatch by indexing into the dataset
		self.batch_num = shared(np.int32(0))
		batch_start = self.batch_num * self.batch_size
		batch_end = batch_start + self.batch_size
		self.batch = self.dataset[batch_start : batch_end,]

		# Define an update that moves the batch window through the dataset
		self.updates = [(self.batch_num, self.batch_num+1)]


	def load_dataset(self, dataset):
		# Load the dataset onto the gpu
		self.dataset.set_value(dataset)
		# Determine the total number of minibatches
		self.num_batches = int(np.ceil(len(dataset) / float(self.batch_size)))
		return self.num_batches


	def get_batch(self):
		return self.batch


	def get_updates(self):
		return self.updates


	def get_num_batches(self):
		return self.num_batches



class NoiseContrastiveTheanoMinibatcher(TheanoMinibatcher):

	def __init__(
		self,
		batch_size=1000,
		noise_ratio=15,
		dtype="float32",
		num_dims=2
	):
		self.batch_size = batch_size
		self.noise_ratio = noise_ratio
		self.dtype = dtype
		self.num_dims = num_dims

		self._setup_batching()


	def _setup_batching(self):

		# Make an empty shared variable that will store the dataset
		# Although empty, we can setup the relationship between the
		# minibatch variable and the full dataset
		self.signal_examples = shared(
			self._initialize_data_container(self.num_dims, self.dtype)
		)
		self.noise_examples = shared(
			self._initialize_data_container(self.num_dims, self.dtype)
		)

		# Make minibatch by indexing into the dataset
		self.batch_num = shared(np.int32(0))

		# Take a sliding minibatch window on the signal_examples
		signal_batch_start = self.batch_num * self.batch_size
		signal_batch_end = signal_batch_start + self.batch_size
		signal_batch = self.signal_examples[signal_batch_start : signal_batch_end,]

		# Take a sliding minibatch window on the noise_examples
		noise_batch_start = self.batch_num * self.batch_size * self.noise_ratio
		noise_batch_end = noise_batch_start + self.batch_size * self.noise_ratio
		noise_batch = self.noise_examples[noise_batch_start : noise_batch_end,]

		# Concatenate the signal and noise minibatch into the full minibatch
		self.batch = T.concatenate((signal_batch, noise_batch))

		# Define an update that moves the batch window through the dataset
		self.updates = [(self.batch_num, self.batch_num+1)]


	def load_dataset(self, signal_examples, noise_examples):
		'''
		Load the dataset onto the GPU.  Determine (and return) the number of
		minibatches.
		'''

		# Reset the internal pointer
		self.reset()

		# Determine the total number of minibatches
		self.num_batches = int(np.ceil(len(signal_examples) / float(self.batch_size)))

		# Check if the dataset divides evenly into batches
		warn_last_batch = False
		expected_len_signal = self.num_batches * self.batch_size
		if expected_len_signal > len(signal_examples):
			warn_last_batch = True

		expected_len_noise = self.num_batches * self.batch_size * self.noise_ratio
		if expected_len_noise > len(noise_examples):
			warn_last_batch = True

		# If dataset doesn't divide evenly into batches, warn the user, and
		# drop the last batch
		if warn_last_batch:
			print 'Warning: incomplete last batch will be ommitted'
			# We ommit the last batch simply by reporting fewer total batches.
			# It is actually up to the caller to only use self.batch_num
			# minibatches.
			self.num_batches -= 1

		# Load the dataset onto the gpu
		self.signal_examples.set_value(signal_examples)
		self.noise_examples.set_value(noise_examples)

		return self.num_batches



import re
import gc
import time
from iterable_queue import IterableQueue
from multiprocessing import Process
from subprocess import check_output
from counter_sampler import CounterSampler
from token_map import UNK
from unigram_dictionary import UnigramDictionary
import numpy as np
import os
import sys


class TokenChooser(object):

	'''
	This choses which context token should be taken given a window
	of +/- K around a query token
	'''

	def __init__(self, K, kernel):
		if not len(kernel) == 2*K:
			raise ValueError(
				'`kernel` must have 2*K entries, one for '
				'each of the elements within the windows of +/- K tokens.'
			)

		self.K = K
		self.kernel = kernel
		self.samplers = {}
		self.indices = range(-K, 0) + range(1, K+1)


	def choose_token(self, idx, length):
		'''
		Randomly choose a token according to the kernel supplied
		in the constructor.  Note that when sampling the context near
		the beginning of a sentence, the left part of the context window
		will be truncated.  Similarly, sampling context near the end of
		a sentence leads to truncation of the right part of the context
		window.  Short sentences lead to truncation on both sides.

		To ensure that samples are returned within the possibly truncated
		window, two values define the actual extent of the context to be
		sampled:

		`idx`: index of the query word within the context.  E.g. if the
			valid context is constrained to a sentence, and the query word
			is the 3rd token in the sentence, idx should be 2 (because
			of 0-based indexing)

		`length`: length of the the context, E.g. If context is
			constrained to a sentence, and sentence is 7 tokens long,
			length should be 7.
		'''

		# If the token is near the edges of the context, then the
		# sampling kernel will be truncated (we can't sample before the
		# firs word in the sentence, or after the last word).
		# Determine the slice indices that define the truncated kernel.
		negative_idx = length - idx
		start = max(0, self.K - idx)
		stop = min(2*self.K, self.K + negative_idx - 1)

		# We make a separate multinomial sampler for each different
		# truncation of the kernel, because they each define a different
		# set of sampling probabilityes.  If we don't have a sampler for
		# this particular kernel shape, make one.
		if not (start, stop) in self.samplers:

			trunc_probabilities = self.kernel[start:stop]
			self.samplers[start,stop] = (
				CounterSampler(trunc_probabilities)
			)

		# Sample from the multinomial sampler for the context of this shape
		outcome_idx = self.samplers[start,stop].sample()

		# Map this into the +/- indexing relative to the query word
		relative_idx = self.indices[outcome_idx + start]

		# And then map this into absolute indexing
		result_idx = relative_idx + idx

		return result_idx


MAX_NUMPY_SEED = 4294967295
def reseed():
	'''
	Makes a hop in the random chain.
	If called before spawning a child processes, it will ensure each child
	generates random numbers independently.  Unlike seeding child randomness
	from an os source of randomness, this is reproducible by starting the 
	parent with the same random seed.
	'''
	np.random.seed(np.random.randint(MAX_NUMPY_SEED))


class DataSetReaderIllegalStateException(Exception):
	'''
	Used if DatasetReader's methods are called in an incorrect order, e.g.
	calling generate_dataset() before calling prepare() on a DatasetReader
	that was not initialized with a UnigramDictionary.
	'''
	pass


def default_parse(filename, verbose=True):
	'''
	Parses input corpus files into a file-format-independent in-memory
	representation.  The output of this function is passed into
	`build_examples` for any processing that is needed, irrespective of
	file format, to generate examples form the stored data.

	INPUTS
	* filename [str]: path to a corpus file to be read

	RETURNS
	* [any]: file-format-independent representation of training data.
	'''
	tokenized_sentences = []
	for line in open(filename):
		tokenized_sentences.append(line.strip().split())
	return tokenized_sentences


class DatasetReader(object):

	def __init__(
		self,
		files=[],
		directories=[],
		skip=[],
		noise_ratio=15,
		t=1e-5,
		num_processes=3,
		unigram_dictionary=None,
		min_frequency=0,
		kernel=[1,2,3,4,5,5,4,3,2,1],
		load_dictionary_dir=None,
		max_queue_size=0,
		macrobatch_size=20000,
		parse=default_parse,
		verbose=True
	):

		# Register parameters to instance namespace
		self.files = files
		self.directories = directories
		self.skip = [re.compile(s) for s in skip]
		self.t = t
		self.noise_ratio = noise_ratio
		self.num_processes = num_processes
		self.kernel = kernel
		self.max_queue_size = max_queue_size
		self.macrobatch_size = macrobatch_size
		self._parse = parse
		self.verbose = verbose
		self.min_frequency = min_frequency

		# If unigram dictionary not supplied, make one
		self.prepared = False
		self.unigram_dictionary = UnigramDictionary()

		# But load dictionary from file if load_dictionary_dir specified.
		if load_dictionary_dir is not None:
			if verbose:
				print 'Loading dictionary from %s...' % load_dictionary_dir
			self.load_dictionary(load_dictionary_dir)

		# Or, if an existing dictionary was passed in, use it
		if unigram_dictionary is not None:
			if verbose:
				print 'A dictionary was supplied'
			self.unigram_dictionary = unigram_dictionary
			self.prune()
			self.prepare = True


	def is_prepared(self):
		'''
		Checks to see whether the dataset reader is ready to generate data.
		Given the simplicity, and that the logic of deciding when 
		self.prepared is True is found elsewhere, this method may seem
		unnecessary.  However, it provides a hook for more complex checking
		in subclasses.
		'''

		if self.prepared:
			return True
		return False


	def parse(self, filename):
		'''
		Delegate to the parse function given to the constructor.
		'''
		return self._parse(filename, self.verbose)


	def check_access(self, save_dir):
		'''
		Test out writing in save_dir.  The processes that generate the data
		to be saved can be really long-running, so we want to find out if 
		there is a simple IOError early!
		'''
		save_dir = os.path.abspath(save_dir)
		path, dirname = os.path.split(save_dir)

		# Make sure that the directory we want exists (make it if not)
		if not os.path.isdir(path):
			raise IOError('%s is not a directory or does not exist' % path)
		if not os.path.exists(save_dir):
			os.mkdir(save_dir)
		elif os.path.isfile(save_dir):
			raise IOError('%s is a file. %' % save_dir)

		# Make sure we can write to the file
		f = open(os.path.join(
			save_dir, '.__test-minibatch-generator-access'
		), 'w')
		f.write('test')
		f.close
		os.remove(os.path.join(
			save_dir, '.__test-minibatch-generator-access' ))


	def generate_filenames(self):
		'''
		Generator that yields all filenames (absolute paths) making up the
		dataset.  (Files are specified to the Minibatcher constructor
		files and / or directories.  All listed files and all files directly
		contained in listed directories will be processed, unless they
		match regex patterns in the optional `skip` list.

		(no INPUTS)

		YIELDS
		* [str]: absolute path to a dataset file
		'''

		# Process the files listed in `files`, unles matches entry in skip
		if self.files is not None:
			for filename in self.files:
				filename = os.path.abspath(filename)

				# Skip files if they match a regex in skip
				if any([s.search(filename) for s in self.skip]):
					continue

				if self.verbose:
					print '\tprocessing', filename

				yield filename

		# Process all the files listed in each directory, unless they
		# match an entry in skip
		if self.directories is not None:
			for dirname in self.directories:
				dirname = os.path.abspath(dirname)

				# Skip directories if they match a regex in skip
				if any([s.search(dirname) for s in self.skip]):
					continue

				for filename in os.listdir(dirname):
					filename = os.path.join(dirname, filename)

					# Only process the *files* under the given directories
					if not os.path.isfile(filename):
						continue

					# Skip files if they match a regex in skip
					if any([s.search(filename) for s in self.skip]):
						continue

					if self.verbose:
						print '\tprocessing', filename

					yield filename


	def numpyify(self, examples):
		'''
		Make an int32-type numpy array, ensuring that, even if the list of
		examples is empty, the array is two-dimensional, with the second
		dimension (i.e. number of columns) being 3.
		'''

		if len(examples) > 0:
			examples = np.array(examples, dtype='int32')
		else:
			examples = np.empty(shape=(0,3), dtype='int32')

		return examples


	# TODO: this should be called 'produce macrobatches'
	def generate_macrobatches(self, filename_iterator):

		'''
		Assembles bunches of examples from the parsed data coming from
		files that were read.  Normally, this function might yield
		individual examples, however, in this case, we need to maintain
		a distinction between the noise- and signal-examples, and to
		keep them in consistent proportions.  So, here, we yield small
		bunches that consist of 1 signal example, and X noise examples,
		where X depends on `self.noise_ratio`.
		'''

		mcbatch_size = self.macrobatch_size
		noise_ratio = self.noise_ratio
		signal_examples = []
		noise_examples = []

		examples = self.generate_examples(filename_iterator)
		for signal_chunk, noise_chunk in examples:

			signal_examples.extend(signal_chunk)
			noise_examples.extend(noise_chunk)

			# Whenever we have enough examples, yield a macrobatch
			while len(signal_examples) > mcbatch_size:
				if self.verbose:
					print 'numpyifying'
				signal_macrobatch = self.numpyify(
					signal_examples[:mcbatch_size])
				noise_macrobatch = self.numpyify(
					noise_examples[:mcbatch_size * noise_ratio])

				if self.verbose:
					print 'no-padding:', len(signal_macrobatch)
				yield signal_macrobatch, noise_macrobatch

				signal_examples = signal_examples[mcbatch_size:]
				noise_examples = noise_examples[mcbatch_size*noise_ratio:]

		# After all files were processed, pad any remaining examples
		# to make up a final macrobatch
		if len(signal_examples) > 0:
			signal_remaining = mcbatch_size - len(signal_examples)
			noise_remaining = (
				mcbatch_size * noise_ratio - len(noise_examples))

			if self.verbose:
				print 'padding and numpyifying'

			padding_row = self.get_padding_row()
			signal_macrobatch = self.numpyify(
				signal_examples + [padding_row] * signal_remaining)
			noise_macrobatch = self.numpyify(
				noise_examples + [padding_row] * noise_remaining)

			if self.verbose:
				print 'padded to length:', len(signal_macrobatch)
			yield signal_macrobatch, noise_macrobatch


	def get_padding_row(self):
		return [UNK,UNK]


	def generate_dataset_serial(self):
		'''
		Generate the dataset from files handed to the constructor.  
		A single process is used to read the files. 
		'''

		# This cannot be called before calling prepare(), unless a prepared
		# UnigramDictionary was passed to the self's constructor
		if not self.is_prepared():
			raise DataSetReaderIllegalStateException(
				"DatasetReader: generate_examples() cannot be called "
				"before prepare() is called unless a prepared "
				"UnigramDictionary has was passed into the DatasetReader's "
				"constructor."
			)

		# Generate the data for each file
		file_iterator = self.generate_filenames()
		macrobatches = self.generate_macrobatches(file_iterator)
		for signal_examples, noise_examples in macrobatches:
			yield signal_examples, noise_examples


	def generate_dataset_worker(self, file_iterator, macrobatch_queue):
		macrobatches = self.generate_macrobatches(file_iterator)
		for signal_examples, noise_examples in macrobatches:
			if self.verbose:
				print 'sending macrobatch to parent process'
			macrobatch_queue.put((signal_examples, noise_examples))

		macrobatch_queue.close()


	def generate_dataset_parallel(self, save_dir=None):
		'''
		Parallel version of generate_dataset_serial.  Each worker is 
		responsible for saving its own part of the dataset to disk, called 
		a macrobatch.  the files are saved at 
		'save_dir/examples/<batch-num>.npz'.
		'''
		# This cannot be called before calling prepare(), unless a prepared
		# UnigramDictionary was passed to the self's constructor

		if not self.is_prepared():
			raise DataSetReaderIllegalStateException(
				"DatasetReader: generate_examples() cannot be called "
				"before prepare() is called unless a prepared "
				"UnigramDictionary has was passed into the "
				"DatasetReader's constructor."
			)

		# We save dataset in the "examples" subdir of the model_dir
		if save_dir is not None:
			examples_dir = os.path.join(save_dir, 'examples')
			# We are willing to create both the save_dir, and the
			# 'examples' subdir, but not their parents
			if not os.path.exists(save_dir):
				os.mkdir(save_dir)
			if not os.path.exists(examples_dir):
				os.mkdir(examples_dir)
		else:
			examples_dir = None

		file_queue = IterableQueue()
		macrobatch_queue = IterableQueue(self.max_queue_size)

		# Put all the filenames on a producer queue
		file_producer = file_queue.get_producer()
		for filename in self.generate_filenames():
			file_producer.put(filename)
		file_producer.close()

		# Start a bunch of worker processes
		for process_num in range(self.num_processes):
			# Hop to a new location in the random-number-generator's state 
			# chain
			reseed()
			# Start child process that generates a portion of the dataset
			args = (
				file_queue.get_consumer(),
				macrobatch_queue.get_producer()
			)
			Process(target=self.generate_dataset_worker, args=args).start()

		# This will receive the macrobatches from all workers
		macrobatch_consumer = macrobatch_queue.get_consumer()

		# Close the iterable queues
		file_queue.close()
		macrobatch_queue.close()

		# Retrieve the macrobatches from the workers, write them to file
		signal_macrobatches = []
		noise_macrobatches = []
		for signal_macrobatch, noise_macrobatch in macrobatch_consumer:

			if self.verbose:
				print 'receiving macrobatch from child process'

			yield signal_macrobatch, noise_macrobatch


	def get_vocab_size(self):
		'''
		Get the size of the vocabulary.  Only makes sense to call this
		after Minibatcher.prepare() has been called, or if an
		existing (pre-filled) UnigramDictionary was loaded, since otherwise
		it would just return 0.

		(no INPUTS)

		OUTPUTS
		* [int]: size of vocabulary (including `UNK`).

		'''
		# Delegate to the underlying UnigramDictionary
		return len(self.unigram_dictionary)


	def load_dictionary(self, load_dir):
		'''
		Loads the unigram_dictionary from files stored in the supplied
		directory.

		INPUTS
		* directory [str]: Path to a directory in which unigram_dictionary
			files are stored.  Unigram dictionary will look for default
			filenames within that directory.

		OUTPUTS
		* [None]
		'''
		# Delegate to the underlying UnigramDictionary
		self.unigram_dictionary.load(os.path.join(
			load_dir, 'dictionary'
		))
		self.prune()

		# It is now possible to call the data generators
		# `generate_dataset_serial()` and `generate_dataset_parallel()`
		self.prepared = True


	def save_dictionary(self, save_dir):
		'''
		Save the unigram_dictionary in the subfolder 'dictionary' beneath
		save_dir, in two files called 'counter-sampler.gz' and 
		'token-map.gz'.  `save_dir` will be created if it doesn't exist.
		'''
		# Make save_dir if it doesn't exist
		if not os.path.exists(save_dir):
			os.mkdir(save_dir)

		# Delegate to the underlying UnigramDictionary
		self.unigram_dictionary.save(os.path.join(
			save_dir, 'dictionary'
		))


	def preparation(self, save_dir, min_frequency=None):

		# Read through the corpus, building the UnigramDictionary
		for filename in self.generate_filenames():
			for tokens in self.parse(filename):
				self.unigram_dictionary.update(tokens)

		self.prepared = True


	def prepare(self, save_dir=None, *args, **kwargs):
		'''
		Used to perform any preparation steps that are needed before
		minibatching can be done.  E.g. assembling a dictionary that
		maps tokens to integers, and determining the total vocabulary size
		of the corpus.  It is assumed that files will need
		to be saved as part of this process, and that they should be
		saved under `save_dir`, with `self.save()` managing the details
		of writing files under `save_dir`.

		INPUTS

		* Note About Inputs *
		the call signature of this method is variable and is
		determined by the call signature of the core
		`self.preparation()` method.  Refer to that method's call
		signature.  Minimally, this method accepts `save_dir`

		* save_dir [str]: path to directory in which preparation files
			should be saved.

		RETURNS
		* [None]
		'''

		# Before doing anything, if we were requested to save the
		# dictionary, make sure we'll be able to do that (fail fast)
		if save_dir is not None:
			self.check_access(save_dir)

		self.preparation(save_dir, *args, **kwargs)

		# Save the dictionary, if requested to do so.
		if save_dir is not None:
			self.save_dictionary(save_dir)

		# Prune the dictionary
		self.prune()



	def prune(self):
		'''
		Exposes the prune function for the underlying UnigramDictionary
		'''
		if self.verbose:
			print (
				'pruning dictionary to eliminate tokens occuring less than '
				'%d times.' % self.min_frequency
			)
		self.unigram_dictionary.prune(self.min_frequency)


	def generate_examples(self, filename_iterator):
		'''
		Using the data of a parsed file, generates examples.  Two kinds of
		examples are generated --- signal and noise.  They are yielded in a
		tuple, along with a flag indicating whether the particular example 
		is a signal, i.e.: (is_signal, example)
		'''

		num_examples = 0
		chooser = TokenChooser(K=len(self.kernel)/2, kernel=self.kernel)
		#sorted_examples = []

		for filename in filename_iterator:

			# Parse the file, then generate a bunch of examples from it
			parsed = self.parse(filename)

			for tokens in parsed:

				# Isolated tokens (e.g. one-word sentences) have no context
				# and can't be used for training.
				if len(tokens) < 2:
					continue

				token_ids = self.unigram_dictionary.get_ids(tokens)

				for query_token_pos, query_token_id in enumerate(token_ids):

					# Possibly discard the token
					if self.do_discard(query_token_id):
						continue

					# Sample a token from the context
					context_token_pos = chooser.choose_token(
						query_token_pos, len(token_ids)
					)
					context_token_id = token_ids[context_token_pos]
					signal_examples = [[query_token_id, context_token_id]]
					num_examples += 1

					noise_examples = self.generate_noise_examples(
						signal_examples)

					num_examples += len(noise_examples)

					yield (signal_examples, noise_examples)


	def generate_noise_examples(self, signal_examples):

		noise_examples = []
		for query_token_id, context_token_id in signal_examples:
			noise_examples.extend([
				[query_token_id, self.unigram_dictionary.sample()]
				for i in range(self.noise_ratio)
			])

		return noise_examples


	#def parsed(self, filename):
	#	tokenized_sentences = []
	#	for line in open(filename):
	#		tokenized_sentences.append(line.strip().split())
	#	return tokenized_sentences


	def make_null_example(self):
		return [UNK, UNK]


	def do_discard(self, token_id):
		'''
		This function helps with downsampling of very common words.
		Returns true when the token should be discarded as a query word
		'''
		probability = self.unigram_dictionary.get_probability(token_id)
		discard_probability = 1 - np.sqrt(self.t/probability)
		do_discard = np.random.uniform() < discard_probability

		return do_discard



from noise_contrast import noise_contrast, get_noise_contrastive_loss
from w2v import word2vec, sigmoid, row_dot, Word2VecEmbedder
from token_map import TokenMap, UNK, SILENT, ERROR
from counter_sampler import CounterSampler, CounterSamplerException
from unigram_dictionary import UnigramDictionary
from dataset_reader import (
	TokenChooser, DatasetReader, DataSetReaderIllegalStateException, reseed
)
from theano_minibatcher import TheanoMinibatcher



from unigram_dictionary import UnigramDictionary
from collections import Counter, defaultdict
from token_map import TokenMap, SILENT, ERROR, UNK
import time
from unittest import main, TestCase
from theano import tensor as T, function, shared
import numpy as np
from w2v import Word2VecEmbedder, word2vec
from noise_contrast import get_noise_contrastive_loss, noise_contrast
from dataset_reader import TokenChooser, DatasetReader, DataSetReaderIllegalStateException
from theano_minibatcher import (
	TheanoMinibatcher, NoiseContrastiveTheanoMinibatcher
)
from counter_sampler import CounterSampler
from lasagne.init import Normal
from lasagne.updates import nesterov_momentum
import re
import os
import shutil
from copy import deepcopy


def sigma(a):
	return 1/(1+np.exp(-a))
usigma = np.vectorize(sigma)


class TestUnigramDictionary(TestCase):
	'''
	Tests that UnigramDictionary properly represents the corpus
	statistics, and that the pruning function works as expected.
	'''

	TOKENS = ['apple', 'pear', 'banana', 'orange']

	# Make a toy corpus with specific token frequencies.
	FREQUENCIES = {
		'apple':4, 'banana':8, 'orange':6,
		'pineapple':3, 'grapefruit':9
	}
	CORPUS = list(Counter(FREQUENCIES).elements())

	def test_sort(self):
		np.random.seed(1)
		np.random.shuffle(self.CORPUS)
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.CORPUS)

		ordered_tokens = ['UNK'] + [
			token for token, count 
			in Counter(self.FREQUENCIES).most_common()
		]
		ordered_counts = [0] + [
			count for token, count 
			in Counter(self.FREQUENCIES).most_common()
		]

		self.assertNotEqual(
			unigram_dictionary.token_map.tokens,
			ordered_tokens
		)
		self.assertNotEqual(
			unigram_dictionary.counter_sampler.counts,
			ordered_counts
		)

		unigram_dictionary.sort()
		
		self.assertEqual(
			unigram_dictionary.token_map.tokens,
			ordered_tokens
		)
		self.assertEqual(
			unigram_dictionary.counter_sampler.counts,
			ordered_counts
		)

		


	def test_remove_compact(self):
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.CORPUS)
		unigram_dictionary.remove('banana')
		unigram_dictionary.remove('pineapple')
		unigram_dictionary.compact()
		
		adjusted_frequencies = deepcopy(self.FREQUENCIES)
		adjusted_frequencies['banana'] = 0
		adjusted_frequencies['pineapple'] = 0
		total = sum(adjusted_frequencies.values())
		reduced_tokens = [
			key for key, val in self.FREQUENCIES.iteritems()
			if val > 0
		]

		for token in reduced_tokens:
			expected_probability = (
				adjusted_frequencies[token] / float(total))
			idx = unigram_dictionary.get_id(token)
			self.assertEqual(
				unigram_dictionary.get_probability(idx), 
				expected_probability
			)

		# Attempting to remove tokens that don't exist is an error
		with self.assertRaises(ValueError):
			unigram_dictionary.remove('fake')

		# Attempting to remove the 'UNK' special token is an error
		with self.assertRaises(ValueError):
			unigram_dictionary.remove('UNK')


	def test_sampling(self):
		'''
		Test basic function of assigning counts, and then sampling from
		The distribution implied by those counts.
		'''
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.CORPUS)

		# Test asking for a single sample (where no shape tuple supplied)
		single_sample = unigram_dictionary.sample()
		self.assertTrue(type(single_sample) is np.int64)

		# Test asking for an array of samples (by passing a shape tuple)
		shape = (2,3,5)
		array_sample = unigram_dictionary.sample(shape)
		self.assertTrue(type(array_sample) is np.ndarray)
		self.assertTrue(array_sample.shape == shape)


	def test_counter_sampler_statistics(self):
		'''
		This tests that the UnigramDictionary really does produce results
		whose statistics match those requested by the counts vector
		'''
		# Seed numpy's random function to make the test reproducible
		np.random.seed(1)

		# Make a sampler with probabilities proportional to counts
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.CORPUS)

		# Draw one hundred thousand samples, then total up the fraction of
		# each outcome obseved
		counter = Counter(unigram_dictionary.sample((100000,)))

		# Make a list of the expected fractions by which each outcome
		# should be observed, in the limit of infinite sample
		total_in_expected = float(len(self.CORPUS))

		tolerance = 0.002
		for idx, found_freq in counter.iteritems():
			found_frac = found_freq / 100000.0
			token = unigram_dictionary.get_token(idx)
			expected_frac = self.FREQUENCIES[token] / total_in_expected
			self.assertTrue(abs(found_frac - expected_frac) < tolerance)


	def test_unigram_dictionary_token_map(self):

		unigram_dictionary = UnigramDictionary(on_unk=SILENT)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that ids are assigned in an auto-incrementing way
			# starting from 1 (0 is reserved for the UNK token)
			self.assertEqual(unigram_dictionary.add(fruit), idx+1)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that idxs are stable and retrievable with
			# UnigramDictionary.get_id()
			self.assertEqual(unigram_dictionary.get_id(fruit), idx+1)

			# Ensure that we can look up the token using the id
			self.assertEqual(unigram_dictionary.get_token(idx+1), fruit)

		# Ensure the unigram_dictionary knows its own length
		self.assertEqual(len(unigram_dictionary), len(self.TOKENS)+1)

		# Asking for ids of non-existent tokens returns the UNK token_id
		self.assertEqual(unigram_dictionary.get_id('no-exist'), 0)

		# Asking for the token at 0 returns 'UNK'
		self.assertEqual(unigram_dictionary.get_token(0), 'UNK')

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			unigram_dictionary.get_token(99)


	def test_raise_error_on_unk(self):
		'''
		If the unigram_dictionary is constructed passing
			on_unk=UnigramDictionary.ERROR
		then calling get_id() or get_ids() will throw a KeyError if one
		of the supplied tokens isn't in the unigram_dictionary.
		(Normally it would return 0, which is a token id reserved for
		'UNK' -- any unknown token).
		'''

		unigram_dictionary = UnigramDictionary(on_unk=ERROR)
		unigram_dictionary.update(self.TOKENS)

		with self.assertRaises(KeyError):
			unigram_dictionary.get_id('no-exist')

		with self.assertRaises(KeyError):
			unigram_dictionary.get_ids(['apple', 'no-exist'])



	def test_plural_functions(self):

		unigram_dictionary = UnigramDictionary(on_unk=SILENT)

		# In these assertions, we offset the expected list of ids by 1
		# because the 0th id in unigram_dictionary is reserved for the UNK
		# token

		# Ensure that update works
		ids = unigram_dictionary.update(self.TOKENS)
		self.assertEqual(ids, range(1, len(self.TOKENS)+1))

		# Ensure that get_ids works
		self.assertEqual(
			unigram_dictionary.get_ids(self.TOKENS),
			range(1, len(self.TOKENS)+1)
		)

		# Ensure that get_tokens works
		self.assertEqual(
			unigram_dictionary.get_tokens(range(1, len(self.TOKENS)+1)),
			self.TOKENS
		)

		# Asking for ids of non-existent tokens raises KeyError
		self.assertEqual(
			unigram_dictionary.get_ids(['apple', 'no-exist']),
			[self.TOKENS.index('apple')+1, 0]
		)

		# Asking for token at 0 returns the 'UNK' token
		self.assertEqual(
			unigram_dictionary.get_tokens([3,0]),
			[self.TOKENS[3-1], 'UNK']
		)

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			unigram_dictionary.get_tokens([1,99])


	def test_save_load(self):

		unigram_dictionary = UnigramDictionary(on_unk=SILENT)
		unigram_dictionary.update(self.CORPUS)
		unigram_dictionary.save('test-data/test-unigram-dictionary')

		unigram_dictionary_copy = UnigramDictionary(on_unk=SILENT)
		unigram_dictionary_copy.load('test-data/test-unigram-dictionary')

		# Test that the mapping from tokens to ids is unchanged
		for token in self.FREQUENCIES:
			self.assertEqual(
				unigram_dictionary.get_id(token),
				unigram_dictionary_copy.get_id(token)
			)

		# Test that the vocabulary size is as expected
		self.assertEqual(
			len(unigram_dictionary_copy),
			len(self.FREQUENCIES)+1
		)

		# Test that the counts for each token are correct
		for token, count in self.FREQUENCIES.items():
			token_id = unigram_dictionary.get_id(token)
			self.assertEqual(
				unigram_dictionary_copy.get_frequency(token_id),
				count
			)

		# Test that the number of tokens is as expected
		self.assertEqual(
			unigram_dictionary_copy.get_num_tokens(),
			sum(self.FREQUENCIES.values())
		)


	def test_pruning(self):


		# Make a unigram dictionary, and populate it with the corpus
		unigram_dictionary = UnigramDictionary()
		unigram_dictionary.update(self.CORPUS)

		# Ensure that the dictionary has correctly encoded the desired
		# information about the corpus.
		for token in unigram_dictionary.token_map.tokens:
			token_id = unigram_dictionary.get_id(token)
			freq = unigram_dictionary.get_frequency(token_id)
			if token == 'UNK':
				self.assertEqual(freq, 0)
			else:
				self.assertEqual(freq, self.FREQUENCIES[token])

		# Check that the dictionary knows the correct number of words.
		# Recall that this is one more than the number of unique words
		# in the corpus, because of the reserved 'UNK' word.
		num_tokens = len(self.FREQUENCIES)+1
		self.assertEqual(len(unigram_dictionary), num_tokens)

		# Prune the dictionary!
		unigram_dictionary.prune(min_frequency=6)

		# Check that two elements were dropped from the dictionary
		# ('apple' and 'pineapple').
		num_tokens -= 2
		self.assertEqual(len(unigram_dictionary), num_tokens)

		# Check that the frequences are as expected.  Counts for
		# apple and pineapple should have been attributed to UNK
		unk_freq = (
			self.FREQUENCIES['apple'] + self.FREQUENCIES['pineapple']
		)
		for token in unigram_dictionary.token_map.tokens:

			# We should not see apple or pineapple
			self.assertTrue(token not in ('apple', 'pineapple'))

			token_id = unigram_dictionary.get_id(token)
			freq = unigram_dictionary.get_frequency(token_id)
			if token == 'UNK':
				self.assertEqual(freq, unk_freq)
			else:
				self.assertEqual(freq, self.FREQUENCIES[token])


class TestTokenChooser(TestCase):
	'''
	Given a list of tokens (usually representing a sentence), and a
	particular query token, the job of the TokenChooser is to sample
	from the nearby tokens, within a window of +/- K tokens.
	It does not yield the sampled token itself, but rather, its index.
	A complicating issue is the fact that the query token might be
	near the beginning or end of the token list (sentence), which needs
	to be accounted for so that the sampled index doesn't produce an
	IndexError
	'''

	def test_mid_sentence(self):
		'''
		Test token chooser's performance given that the query token
		is far from the edges of the context
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "sufficiently" as the query word
		sentence = (
			'this is a test sentence that is sufficiently long to '
			'enable testing various edge cases'
		).split()
		query_idx = sentence.index('sufficiently')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples))
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  So the acceptable
		# relative indices are:
		expected_relative_idxs = range(-5, 0) + range(1, 6)

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		total = float(sum(counts))
		expected_frequencies = dict(
			[(idx, c / total) for idx, c in zip(expected_idxs, counts)]
		)

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with
		# approximately the frequencies expected
		tolerance = 0.002
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)


	def test_near_beginning(self):
		'''
		Test token chooser's performance given that the query token
		is close to the beginning of the context
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "a" as the query word
		sentence = (
			'this is a test sentence that is sufficiently long to '
			'enable testing various edge cases'
		).split()
		query_idx = sentence.index('a')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples))
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  But because the
		# query token is near the start of the context, there are actually
		# only two tokens available before it
		expected_relative_idxs = [-2, -1] + range(1, 6)

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		relative_frequencies = [4,5,5,4,3,2,1]
		total = float(sum(relative_frequencies))
		expected_frequencies = dict([
			(idx, c / total)
			for idx, c in zip(expected_idxs, relative_frequencies)
		])

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with
		# approximately the frequencies expected
		tolerance = 0.003
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)


	def test_near_end(self):
		'''
		Test token chooser's performance given that the query token
		is close to the end of the context
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "cases" as the query word
		sentence = (
			'this is a test sentence that is sufficiently long to '
			'enable testing various edge cases'
		).split()
		query_idx = sentence.index('cases')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples))
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  But because the
		# query token is near the start of the context, there are actually
		# only two tokens available before it
		expected_relative_idxs = range(-5, 0)

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		relative_frequencies = [1,2,3,4,5]
		total = float(sum(relative_frequencies))
		expected_frequencies = dict([
			(idx, c / total)
			for idx, c in zip(expected_idxs, relative_frequencies)
		])

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with
		# approximately the frequencies expected
		tolerance = 0.003
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)


	def test_short_context(self):
		'''
		Test token chooser's performance given that context is short
		'''

		# Seed randomness for reproducibility
		np.random.seed(1)

		# Set up some testing data.  We'll sample from this sentence
		# using the word "This" as the query word
		sentence = 'This is short'.split()
		query_idx = sentence.index('This')
		context_length = len(sentence)

		# Sample from +/- 5 words
		K = 5
		# Weight the probabilities like this (higher probability of
		# sampling near the query word itself).
		counts = [1,2,3,4,5,5,4,3,2,1]

		# Make the chooser
		chooser = TokenChooser(K, counts)

		# Sample many times so we can test the sample's statistics
		num_samples = 100000
		found_counts = Counter([
			chooser.choose_token(query_idx, context_length)
			for s in range(num_samples)
		])

		# Convert counts into frequencies
		found_frequencies = dict([
			(idx, c / float(num_samples))
			for idx, c in found_counts.items()
		])

		# Assemble the expected frequencies.  First, the acceptable
		# outcomes are tokens that are within +/- 5.  But because the
		# query token is near the start of the context, there are actually
		# only two tokens available before it
		expected_relative_idxs = [1,2]

		# And so the expected absolute indices are:
		expected_idxs = [
			rel + query_idx
			for rel in expected_relative_idxs
		]

		# Calculate the expected frequency that each index should appear
		# in the sample
		relative_frequencies = [5,4]
		total = float(sum(relative_frequencies))
		expected_frequencies = dict([
			(idx, c / total)
			for idx, c in zip(expected_idxs, relative_frequencies)
		])

		# First, make sure that the sample indices are the ones expected
		self.assertEqual(
			set(found_frequencies.keys()),
			set(expected_frequencies.keys())
		)

		# Then check that the indices in the sample arise with
		# approximately the frequencies expected
		tolerance = 0.003
		for idx in expected_frequencies:
			diff = abs(expected_frequencies[idx] - found_frequencies[idx])
			self.assertTrue(diff < tolerance)



class TestCounterSampler(TestCase):

	def test_sampling(self):
		'''
		Test basic function of assigning counts, and then sampling from
		The distribution implied by those counts.
		'''
		counts = range(1,6)
		counter_sampler = CounterSampler()
		counter_sampler.update(counts)

		# Test asking for a single sample (where no shape tuple supplied)
		single_sample = counter_sampler.sample()
		self.assertTrue(type(single_sample) is np.int64)

		# Test asking for an array of samples (by passing a shape tuple)
		shape = (2,3,5)
		array_sample = counter_sampler.sample(shape)
		self.assertTrue(type(array_sample) is np.ndarray)
		self.assertTrue(array_sample.shape == shape)


	def test_add_function(self):
		'''
		Make sure that the add function is working correctly.
		CounterSampler stores counts as list, wherein the value at
		position i of the list encodes the number of counts seen for
		outcome i.

		Counts are added by passing the outcome's index into
		CounterSampler.add()
		which leads to position i of the counts list to be incremented.
		If position i doesn't exist, it is created.  If the counts list
		had only j elements before, and a count is added for position
		i, with i much greater than j, then many elements are created
		between i and j, and are provisionally initialized with zero
		counts.

		Ensure that is done properly
		'''

		counter_sampler = CounterSampler()
		self.assertEqual(counter_sampler.counts, [])

		outcome_to_add = 6
		counter_sampler.add(outcome_to_add)
		expected_counts = [0]*(outcome_to_add) + [1]
		self.assertEqual(counter_sampler.counts, expected_counts)

		# Now ensure the underlying sampler can tolerate a counts list
		# containing zeros, and that the sampling statistics is as
		# expected.  We expect that the only outcome that should turn up
		# is outcome 6, since it has all the probability mass.  Check that.
		counter = Counter(counter_sampler.sample((100000,))) # should be
															 # all 6's
		total = float(sum(counter.values()))
		found_normalized = [
			counter[i] / total for i in range(outcome_to_add+1)
		]

		# Make an list of the expected fractions by which each outcome
		# should be observed, in the limit of infinite sample
		expected_normalized = expected_counts

		# Check if each outcome was observed with a fraction that is within
		# 0.005 of the expected fraction
		self.assertEqual(found_normalized, expected_normalized)


	def test_counter_sampler_statistics(self):
		'''
		This tests that the sampler really does produce results whose
		statistics match those requested by the counts vector
		'''
		# Seed numpy's random function to make the test reproducible
		np.random.seed(1)

		# Make a sampler with probabilities proportional to counts
		counts = range(1,6)
		counter_sampler = CounterSampler()
		for outcome, count in enumerate(counts):
			counter_sampler.update([outcome]*count)

		# Draw one hundred thousand samples, then total up the fraction of
		# each outcome obseved
		counter = Counter(counter_sampler.sample((100000,)))
		total = float(sum(counter.values()))
		found_normalized = [
			counter[i] / total for i in range(len(counts))
		]

		# Make an list of the expected fractions by which each outcome
		# should be observed, in the limit of infinite sample
		total_in_expected = float(sum(counts))
		expected_normalized = [
			c / total_in_expected for c in counts
		]

		# Check if each outcome was observed with a fraction that is within
		# 0.005 of the expected fraction
		close = [
			abs(f - e) < 0.005
			for f,e in zip(found_normalized, expected_normalized)
		]
		self.assertTrue(all(close))


	def test_save_load(self):

		fname = 'test-data/test-counter-sampler/test-counter-sampler.gz'

		# Make a sampler with probabilities proportional to counts
		counts = range(1,6)
		counter_sampler = CounterSampler()
		for outcome, count in enumerate(counts):
			counter_sampler.update([outcome]*count)

		counter_sampler.save(fname)

		new_counter_sampler = CounterSampler()
		new_counter_sampler.load(fname)
		self.assertEqual(new_counter_sampler.counts, counts)


class TestTokenMap(TestCase):

	TOKENS = ['apple', 'pear', 'banana', 'orange']

	def test_token_map(self):

		token_map = TokenMap(on_unk=SILENT)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that ids are assigned in an auto-incrementing way
			# starting from 1 (0 is reserved for the UNK token)
			self.assertEqual(token_map.add(fruit), idx+1)

		for idx, fruit in enumerate(self.TOKENS):
			# Ensure that idxs are stable and retrievable with
			# TokenMap.get_id()
			self.assertEqual(token_map.get_id(fruit), idx+1)

			# Ensure that we can look up the token using the id
			self.assertEqual(token_map.get_token(idx+1), fruit)

		# Ensure the token_map knows its own length
		self.assertEqual(len(token_map), len(self.TOKENS)+1)

		# Asking for ids of non-existent tokens returns the UNK token_id
		self.assertEqual(token_map.get_id('no-exist'), 0)

		# Asking for the token at 0 returns 'UNK'
		self.assertEqual(token_map.get_token(0), 'UNK')

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			token_map.get_token(99)


	def test_raise_error_on_unk(self):
		'''
		If the token_map is constructed passing
			on_unk=TokenMap.ERROR
		then calling get_id() or get_ids() will throw a KeyError if one
		of the supplied tokens isn't in the token_map.  (Normally it
		would return 0, which is a token id reserved for 'UNK' -- any
		unknown token).
		'''

		token_map = TokenMap(on_unk=ERROR)
		token_map.update(self.TOKENS)

		with self.assertRaises(KeyError):
			token_map.get_id('no-exist')

		with self.assertRaises(KeyError):
			token_map.get_ids(['apple', 'no-exist'])


	def test_token_map_plural_functions(self):

		token_map = TokenMap(on_unk=SILENT)

		# In these assertions, we offset the expected list of ids by
		# 1 because the 0th id in token_map is reserved for the UNK
		# token

		# Ensure that update works
		ids = token_map.update(self.TOKENS)
		self.assertEqual(ids, range(1, len(self.TOKENS)+1))

		# Ensure that get_ids works
		self.assertEqual(
			token_map.get_ids(self.TOKENS),
			range(1, len(self.TOKENS)+1)
		)

		# Ensure that get_tokens works
		self.assertEqual(
			token_map.get_tokens(range(1, len(self.TOKENS)+1)),
			self.TOKENS
		)

		# Asking for ids of non-existent tokens raises KeyError
		self.assertEqual(
			token_map.get_ids(['apple', 'no-exist']),
			[self.TOKENS.index('apple')+1, 0]
		)

		# Asking for token at 0 returns the 'UNK' token
		self.assertEqual(
			token_map.get_tokens([3,0]),
			[self.TOKENS[3-1], 'UNK']
		)

		# Asking for token at non-existent idx raises IndexError
		with self.assertRaises(IndexError):
			token_map.get_tokens([1,99])


	def test_save_load(self):
		token_map = TokenMap(on_unk=SILENT)
		token_map.update(self.TOKENS)
		token_map.save('test-data/test-token-map/test-token-map.gz')

		token_map_copy = TokenMap(on_unk=SILENT)
		token_map_copy.load(
			'test-data/test-token-map/test-token-map.gz'
		)
		self.assertEqual(
			token_map_copy.get_ids(self.TOKENS),
			range(1, len(self.TOKENS)+1)
		)
		self.assertEqual(len(token_map_copy), len(self.TOKENS)+1)



class TestNoiseContrast(TestCase):

	def test_noise_contrast(self):

		signal_input = T.dvector()
		noise_input = T.dmatrix()
		loss = noise_contrast(signal_input, noise_input)

		f = function([signal_input, noise_input], loss)

		test_signal = np.array([.6, .7, .8])
		test_noise = np.array([[.2, .3, .1],[.5, .6, .7]])
		test_loss = f(test_signal, test_noise)

		expected_objective = (
			np.log(test_signal).sum() + np.log(1-test_noise).sum()
		)
		expected_loss = -expected_objective / float(len(test_signal))

		self.assertAlmostEqual(test_loss, expected_loss)


class TestDataReader(TestCase):

	#TODO: add a test for saving and loading the dictionary and for 
	def setUp(self):

		# Define some parameters to be used in construction
		# Minibatcher
		self.files = [
			'test-data/test-corpus/003.tsv',
			'test-data/test-corpus/004.tsv'
		]
		self.batch_size = 5
		self.macrobatch_size = 5
		self.noise_ratio = 15
		self.num_example_generators = 3
		self.t = 0.03

		self.dataset_reader_with_discard = DatasetReader(
			files=self.files,
			noise_ratio = self.noise_ratio,
			t=self.t,
			macrobatch_size=self.macrobatch_size,
			num_processes=3,
			verbose=False
		)

		self.dataset_reader_no_discard = DatasetReader(
			files=self.files,
			macrobatch_size=self.macrobatch_size,
			noise_ratio = self.noise_ratio,
			t=1.0,
			num_processes=3,
			verbose=False
		)


	def test_prune(self):
		save_dir = 'test-data/test-dataset-reader'
		if os.path.exists(save_dir):
			shutil.rmtree(save_dir)

		files = [
			'test-data/test-corpus/003.tsv',
			'test-data/test-corpus/004.tsv'
		]

		# First make a dataset reader with no min frequency
		reader = DatasetReader(
			files=files,
			min_frequency=0,
			verbose=False
		)
		reader.prepare(save_dir=save_dir)

		# There should be 303 tokens in the dictionary
		self.assertEqual(reader.get_vocab_size(), 303)

		# Now make a dataset reader with a min frequency of 10
		reader = DatasetReader(
			files=files,
			min_frequency=10,
			verbose=False
		)
		reader.prepare()

		# There should be only 7 tokens in dictionary (others got pruned)
		self.assertEqual(reader.get_vocab_size(), 7)

		# Now try making a dataset reader and loading from file,
		# again enforcing the in frequency
		reader = DatasetReader(
			files=files,
			min_frequency=10,
			load_dictionary_dir=save_dir,
			verbose=False
		)
		self.assertEqual(reader.get_vocab_size(), 7)

		# Now try making a dataset reader and loading the dictionary
		# manually
		reader = DatasetReader(
			files=files,
			min_frequency=10,
			verbose=False
		)
		reader.load_dictionary(save_dir)
		self.assertEqual(reader.get_vocab_size(), 7)

		# Now try passing the dictionary into the DatasetReader
		dictionary = UnigramDictionary()
		dictionary.load(os.path.join(save_dir, 'dictionary'))
		reader = DatasetReader(
			files=files,
			min_frequency=10,
			unigram_dictionary=dictionary,
			verbose=False
		)
		self.assertEqual(reader.get_vocab_size(), 7)



	def test_illegal_state_exception(self):
		'''
		Calling generate_dataset_parallel() on a DatasetReader before 
		calling prepare() should raise DataSetReaderIllegalStateException.
		'''

		reader = self.dataset_reader_with_discard
		with self.assertRaises(DataSetReaderIllegalStateException):
			iterator = reader.generate_dataset_parallel()
			iterator.next()

		with self.assertRaises(DataSetReaderIllegalStateException):
			iterator = reader.generate_dataset_serial()
			iterator.next()


	def test_token_discarding(self):
		'''
		Make sure that token discarding is occurring as expected.  We build
		the test case around the token "the", because its the most common 
		token, and so is most reliably discarded.  We compare the number of
		signal examples whose query word is "the" to the number of 
		occurrences of "the" in the test corpus, and check that this 
		fraction is close to that expected based on the prescribed 
		probability of discarding.
		'''
		# Ensure reproducibility in this stochastic test
		np.random.seed(1)
		reader = self.dataset_reader_with_discard
		reader.prepare()

		# Count the number of times "the" appears
		num_the_tokens = 0
		the_pattern = re.compile(r'\bthe\b')
		for filename in self.files:
			text = open(filename).read()
			the_matches = the_pattern.findall(text)
			num_the_tokens += len(the_matches)

		# Repeatedly generate the dataset, with discarding, and keep track 
		# of how many times "the" is included as a query word
		num_replicates = 5
		signal_query_frequencies = []
		for i in range(num_replicates):
			signal_query_freq, noise_query_freq = (
				self.get_the_query_frequency())
			signal_query_frequencies.append(signal_query_freq)
			# The ratio of noise and signal examples should not be affected
			# by discarding.
			self.assertEqual(
				signal_query_freq*self.noise_ratio, noise_query_freq)

		# Take the average frequency with which "the" arose accross 
		# replicates.
		signal_query_frequency = np.mean(signal_query_frequencies)

		# We expect "the" to arise in signal queries less often than it 
		# actually arises in the text, based on the probability of 
		# discarding
		tolerance = 0.05
		the_token = reader.unigram_dictionary.get_id('the')
		the_frequency = reader.unigram_dictionary.get_probability(
			the_token)
		expected_keep_ratio = np.sqrt(self.t/the_frequency)
		actual_keep_ratio = signal_query_frequency / float(num_the_tokens)
		self.assertTrue(
			abs(actual_keep_ratio - expected_keep_ratio) < tolerance
		)


	def get_the_query_frequency(self):
		# Generate the dataset, and see how many times "the" appears as a 
		# query word.
		reader = self.dataset_reader_with_discard
		dataset = reader.generate_dataset_parallel()
		seen_noise_queries = Counter()
		num_the_signals = 0
		num_the_noises = 0
		the_token = reader.unigram_dictionary.get_id('the')
		dataset_iterator = self.iterate_dataset(dataset)
		for signal_examples, noise_examples in dataset_iterator:

			# Count number of times "the" occurs as query in 
			# signal_examples
			signal_queries = signal_examples[:,0]
			num_the_signals += sum([
				t == the_token for t in signal_queries
			])

			# Count number of times "the" occurs as query in noise_examples
			noise_queries = noise_examples[:,0]
			num_the_noises += sum([t == the_token for t in noise_queries])

		return num_the_signals, num_the_noises


	def test_dataset_composition(self):
		'''
		Make sure that the minibatches are the correct size, that
		signal query- and contexts-words are always within 5 tokens of
		one another and come from the same sentence.
		'''
		# Ensure reproducibility in this stochastic test
		np.random.seed(1)

		reader = self.dataset_reader_no_discard
		reader.prepare()

		# Iterate through the corpus, noting what tokens arise within
		# one another's contexts.  Build a lookup table, indicating the set
		# of "legal pairs" -- tokens that arose in one another's context.
		legal_pairs = defaultdict(set)
		# We'll also keep track of the query words in the signal examples
		# To make sure that noise examples are also made for them
		expected_noise_queries = Counter()
		d = reader.unigram_dictionary
		for filename in self.files:
			for tokens in reader.parse(filename):
				token_ids = d.get_ids(tokens)
				for i, token_id in enumerate(token_ids):
					low = max(0, i-5)
					legal_context = token_ids[low:i] + token_ids[i+1:i+6]
					legal_pairs[token_id].update(legal_context)
					# Every time a token appears we expect noise_ratio 
					# noise examples for it
					expected_noise_queries[token_id] += self.noise_ratio

		# finally, and the pair (UNK, UNK), which is used to pad data
		legal_pairs[UNK] = set([UNK])

		reader.prepare()
		dataset = reader.generate_dataset_parallel()
		seen_noise_queries = Counter()
		dataset_iterator = self.iterate_dataset(dataset)
		for signal_examples, noise_examples in dataset_iterator:

			# Keep track of how many times each token appears as the query 
			# in a noise example
			noise_queries = noise_examples[:,0]
			seen_noise_queries.update(noise_queries)

			# Ensure that all of the signal examples are actually valid
			# samples from the corpus
			for query, context in signal_examples:
				self.assertTrue(context in legal_pairs[query])

		# Ensure that we got the expected number of appearances of tokens 
		# in noise examples.  But since we don't care about the number of 
		# times that UNK appears (since it is used for padding), we remove 
		# it from observed counts first
		del seen_noise_queries[UNK]
		keys = set(
			seen_noise_queries.keys() + expected_noise_queries.keys())
		for key in keys:
			self.assertEqual(
				seen_noise_queries[key], expected_noise_queries[key])


	def iterate_dataset(self, dataset):
		'''
		Given a dataset, separate out successive signal and noise examples,
		and iterate through them
		'''

		batch_size = self.batch_size
		noise_batch_size = self.batch_size * self.noise_ratio

		for signal_macrobatch, noise_macrobatch in dataset:
			num_batches = len(signal_macrobatch) / self.batch_size
			for pointer in range(num_batches):

				signal_start = pointer * batch_size
				signal_examples = signal_macrobatch[
					signal_start : signal_start + batch_size
				]
				noise_start = pointer * noise_batch_size
				noise_examples = noise_macrobatch[
					noise_start : noise_start + noise_batch_size
				]

				yield signal_examples, noise_examples


	def test_prepare(self):
		'''
		Check that DatasetReader.prepare() properly makes a
		UnigramDictionary that reflects the corpus.
		'''
		reader = self.dataset_reader_with_discard
		reader.prepare()
		d = reader.unigram_dictionary

		# Make sure that all of the expected tokens are found in the
		# unigram_dictionary, and that their frequency in the is correct.
		tokens = []
		for filename in self.files:
			for add_tokens in reader.parse(filename):
				tokens.extend(add_tokens)

		counts = Counter(tokens)
		for token in tokens:
			token_id = d.get_id(token)
			count = d.get_frequency(token_id)
			self.assertEqual(count, counts[token])


	def test_noise_uniqueness(self):
		'''
		A bug was discovered whereby, when batches are yielded via the
		asynchronous method (i.e. via the iterator returned by
		Minibatcher.get_async_batch_iterator()), different example
		generating processes will generate the same sequence of noise
		contexts, because each example generating process inherits the
		same numpy random number generating state.  This test exposes the
		bug, so that it could be fixed.
		'''

		# Ensure reproducibility in this stochastic test
		np.random.seed(1)

		reader = self.dataset_reader_no_discard
		reader.prepare()
		dataset = reader.generate_dataset_parallel()

		# Iterate through all examples in the dataset.  Look at each of
		# the noise examples generated.  Each query word should have
		# noise_ratio number of noise contexts generated.  No two
		# query words should get the same sequence of noise sequences,
		# which is what we want to test.  Collect the sequence of noise
		# contexts for each query word, and check they are unique
		noise_sequences = defaultdict(list)
		last_query = None
		dataset_iterator = self.iterate_dataset(dataset)
		for signal_examples, noise_examples in dataset_iterator:
			for row in noise_examples:
				query, context = row
				# We don't care about noise generated for UNK queries: they
				# will certainly repeat because UNK is used for padding
				if query == UNK:
					continue
				noise_sequences[query].append(context)


		# Now iterate through all of the accumulated noise sequences, and
		# ensure that we don't see any duplicates
		seen_noise_sequences = set()
		for key in noise_sequences:

			# The full sequence involves noise sequences for multiple 
			# instances of the query.  Divide it into chunks, where each is
			# the noise sequence for an individual instance
			full_sequence = noise_sequences[key]
			instance_sequences = [
				tuple(full_sequence[i:i+self.noise_ratio])
				for i in xrange(0, len(full_sequence), self.noise_ratio)
			]

			# Check if any of these sequences happened already, and add it 
			# to the list of seen noise sequences
			for sequence in instance_sequences:
				self.assertFalse(sequence in seen_noise_sequences)
				seen_noise_sequences.add(sequence)


#	TODO: saving and loading has been disabled -- it doesn't seem to 
# 	make sense under generator-based dataset iteration
#
#	def test_save_load_data_serial(self):
#		self.save_load_data(True)
#
#
#	def test_save_load_data_parallel(self):
#		self.save_load_data(False)
#
#
#	def save_load_data(self, use_serial):
#		'''
#		Test that save and load actually write and read the 
#		Word2VecEmbedder's parameters, by testing that a saved and then 
#		then loaded model yields the same model as the original.
#		'''
#
#		# Clear old testing data, and make sure dir for testing data exists
#		directory = 'test-data/test-dataset-reader'
#		if os.path.exists(directory):
#			shutil.rmtree(directory)
#		os.mkdir(directory)
#
#		# Make a dataset reader, and use two files as the dataset.  Use no
#		# token discarding.
#		reader = DatasetReader(
#			files=[
#				'test-data/test-corpus/numbers-med1.txt',
#				'test-data/test-corpus/numbers-med2.txt'
#			],
#			t=1, num_processes = 2,
#			verbose=False
#		)
#		reader.prepare()
#
#		# Generate the dataset, and pass in `directory` as the location to 
#		# save the generated dataset
#		if use_serial:
#			reader.generate_dataset_serial(directory)
#		else:
#			reader.generate_dataset_parallel(directory)
#
#		# Reload the saved dataset from file
#		new_reader = DatasetReader()
#		new_reader.load_data(directory)
#
#		# Ensure the generated dataset, and the one relaoded from file 
#		# are equal.
#		self.assertTrue(
#			np.array_equal(new_reader.examples, reader.examples))


class TestMinibatcher(TestCase):

	def setUp(self):

		# Define some parameters to be used in construction
		# Minibatcher
		self.files = [
			'test-data/test-corpus/003.tsv',
			'test-data/test-corpus/004.tsv'
		]
		self.batch_size = 5
		self.macrobatch_size = 20
		self.noise_ratio = 15
		self.num_example_generators = 3
		self.t = 0.03

		self.dataset_reader_with_discard = DatasetReader(
			files=self.files,
			noise_ratio = self.noise_ratio,
			t=self.t,
			num_processes=3,
			macrobatch_size=self.macrobatch_size,
			verbose=False
		)

		self.dataset_reader_no_discard = DatasetReader(
			files=self.files,
			noise_ratio = self.noise_ratio,
			t=1.0,
			num_processes=3,
			macrobatch_size=self.macrobatch_size,
			verbose=False
		)


	def test_dataset_composition(self):
		'''
		Make sure that the minibatches are the correct size, that
		signal query- and contexts-words are always within 5 tokens of
		one another and come from the same sentence.
		'''
		# Ensure reproducibility in this stochastic test
		np.random.seed(1)

		reader = self.dataset_reader_no_discard
		reader.prepare()

		# Iterate through the corpus, noting what tokens arise within
		# one another's contexts.  Build a lookup table, indicating the set
		# of "legal pairs" -- tokens that arose in one another's context.
		legal_pairs = defaultdict(set)
		# We'll also keep track of the query words in the signal examples
		# To make sure that noise examples are also made for them
		expected_noise_queries = Counter()
		d = reader.unigram_dictionary
		for filename in self.files:
			for tokens in reader.parse(filename):
				token_ids = d.get_ids(tokens)
				for i, token_id in enumerate(token_ids):
					low = max(0, i-5)
					legal_context = token_ids[low:i] + token_ids[i+1:i+6]
					legal_pairs[token_id].update(legal_context)
					# Every time a token appears we expect noise_ratio 
					# noise examples for it
					expected_noise_queries[token_id] += self.noise_ratio

		# finally, and the pair (UNK, UNK), which is used to pad data
		legal_pairs[UNK] = set([UNK])

		# Make a minibatcher
 		minibatcher = NoiseContrastiveTheanoMinibatcher(
			batch_size=self.batch_size, 
			noise_ratio=self.noise_ratio, 
			dtype='int32',
			num_dims=2
		)
		symbolic_batch = minibatcher.get_batch()
		updates = minibatcher.get_updates()



		# Mock a theano training function.  The function simply returns the
		# current minibatch, so we can check that the minibatches have the
		# expected composition
		f = function([], symbolic_batch, updates=updates)

		seen_noise_queries = Counter()
		for macrobatch in reader.generate_dataset_parallel():
			signal_macrobatch, noise_macrobatch = macrobatch

			# Load the dataset
			num_batches = minibatcher.load_dataset(
				signal_macrobatch, noise_macrobatch
			)

			# Iterate through the minibatches, to check that they have 
			# expected composition
			for i in range(num_batches):

				# Get the batch, and split it into the noise and signal 
				# parts
				batch = f()
				signal_examples = batch[0:self.batch_size, ]
				noise_examples = batch[
					self.batch_size:self.batch_size*(self.noise_ratio+1), ]

				# Keep track of how many times each token appears as the 
				# query in a noise example
				noise_queries = noise_examples[:,0]
				seen_noise_queries.update(noise_queries)

				# Ensure that all of the signal examples are actually valid
				# samples from the corpus
				for query, context in signal_examples:
					self.assertTrue(context in legal_pairs[query])

		# Ensure that we got the expected number of appearances of tokens 
		# in noise examples.  But since we don't care about the number of 
		# times that UNK appears (since it is used for padding), we remove 
		# it from observed counts first
		del seen_noise_queries[UNK]
		keys = set(
			seen_noise_queries.keys() + expected_noise_queries.keys()
		)
		for key in keys:
			self.assertEqual(
				seen_noise_queries[key], expected_noise_queries[key])


	def test_symbolic_minibatches(self):
		'''
		Test that the symbolic minibatching mechanism yields the expected
		batches when used in a compiled theno function.
		'''
		# Ensure reproducibility in this stochastic test
		np.random.seed(1)

		# Make a TheanoMinibatcher; get the symbolic minibatch and updates
		batch_size = 5
		num_batches = 10
		num_examples = batch_size * num_batches
		dtype = 'int32'
		num_dims = 2
		batcher = TheanoMinibatcher(batch_size, dtype, num_dims)
		symbolic_batch = batcher.get_batch()
		updates = batcher.get_updates()

		# Mock a training function.  It just returns the current batch.
		f = function([], symbolic_batch, updates=updates)

		# Load a mock dataset
		dataset = np.array([[i,i] for i in range(num_examples)], dtype=dtype)
		actual_num_batches = batcher.load_dataset(dataset)

		# Check that we have the expected number of batches
		self.assertEqual(actual_num_batches, num_batches)

		for batch_num in range(actual_num_batches):
			batch = f()
			start, stop = batch_size*batch_num, batch_size*(batch_num+1)
			expected_batch = np.array([[i,i] for i in range(start, stop)], dtype=dtype)
			self.assertTrue(np.array_equal(batch, expected_batch))


	def test_data_container_initialization(self):

		# Make a minibatcher so we can test its _initialize_data_container method
		batcher = TheanoMinibatcher()

		# Test the method at various parameter values
		float32_1 = batcher._initialize_data_container(1,'float32')
		uint32_2 = batcher._initialize_data_container(2,'uint32')
		float64_3 = batcher._initialize_data_container(3,'float64')

		# Mock the expected results
		expected_float32_1 = np.array([],dtype='float32')
		expected_uint32_2 = np.array([[]],dtype='uint32')
		expected_float64_3 = np.array([[[]]],dtype='float64')

		# Ensure we got the expected results
		self.assertTrue(np.array_equal(float32_1, expected_float32_1))
		self.assertTrue(np.array_equal(uint32_2, expected_uint32_2))
		self.assertTrue(np.array_equal(float64_3, expected_float64_3))

		# The number of dimensions in a loaded dataset needs to match the number
		# of dimensions given to the TheanoMinibatcher constructor
		batcher = TheanoMinibatcher(num_dims=3, dtype='float32')
		# This works
		batcher.load_dataset(np.array([[[1.0,2.0]]], dtype='float32'))
		# But not this
		with self.assertRaises(TypeError):
			batcher.load_dataset(np.array([[1.0,2.0]], dtype='float32'))

		# Trying to initialize a 0-dimensional dataset should raise a ValueError
		with self.assertRaises(ValueError):
			int32_0 = batcher._initialize_data_container(0,'int32')


	def test_symbolic_minibatches_more_dims(self):
		'''
		Test that the symbolic minibatching mechanism yields the expected
		batches when used in a compiled theno function, and specifically test
		using a non-default number of dimensions in the dataset.
		'''
		# Ensure reproducibility in this stochastic test
		np.random.seed(1)

		# Make a TheanoMinibatcher; get the symbolic minibatch and updates
		batch_size = 5
		num_batches = 10
		num_examples = batch_size * num_batches
		dtype = 'int32'
		num_dims = 3
		batcher = TheanoMinibatcher(batch_size, dtype, num_dims)
		symbolic_batch = batcher.get_batch()
		updates = batcher.get_updates()

		# Mock a training function.  It just returns the current batch.
		f = function([], symbolic_batch, updates=updates)

		# Load a mock dataset
		dataset = np.array([ [[i,i,i] for j in range(2)] for i in range(num_examples)], dtype=dtype)
		actual_num_batches = batcher.load_dataset(dataset)

		# Check that we have the expected number of batches
		self.assertEqual(actual_num_batches, num_batches)

		for batch_num in range(actual_num_batches):
			batch = f()
			start, stop = batch_size*batch_num, batch_size*(batch_num+1)
			expected_batch = np.array(
				[[ [i,i,i] for j in range(2)] for i in range(start, stop)],
				dtype=dtype
			)
			self.assertTrue(np.array_equal(batch, expected_batch))



class TestWord2VecOnCorpus(TestCase):
	'''
	This tests the Word2Vec end-to-end functionality applied to a text
	corpus.
	'''

	def test_word2vec_does_pruning(self):

		files = [
			'test-data/test-corpus/003.tsv',
			'test-data/test-corpus/004.tsv'
		]

		word2vec_embedder, reader = word2vec(
			files=files,
			min_frequency=0,
			num_epochs=1,
			batch_size=int(1e2),
			macrobatch_size=int(1e5),
			t=1,
			num_embedding_dimensions=5,
			verbose=False
		)
		self.assertEqual(reader.get_vocab_size(), 303)

		word2vec_embedder, reader = word2vec(
			files=files,
			min_frequency=10,
			num_epochs=1,
			batch_size=int(1e2),
			macrobatch_size=int(1e5),
			t=1,
			num_embedding_dimensions=5,
			verbose=False
		)
		self.assertEqual(reader.get_vocab_size(), 7)


	def test_word2vec_on_corpus(self):

		# Seed randomness to make the test reproducible
		np.random.seed(1)

		word2vec_embedder, reader = word2vec(
			files=['test-data/test-corpus/numbers-long.txt'],
			num_epochs=1,
			batch_size=int(1e2),
			macrobatch_size=int(1e5),
			t=1,
			num_embedding_dimensions=5,
			verbose=False
		)

		W, C = word2vec_embedder.get_param_values()
		dots = usigma(np.dot(W,C.T))

		# Based on the construction of the corpus, the following
		# context embeddings should match the query at right and be
		# the highest value in the product of the embedding matrices
		# Note that token 0 is reserved for UNK.  It's embedding stays
		# near the randomly initialized value, tending to yield of 0.5
		# which is high overall, so it turns up as a "good match" to any
		# word
		expected_tops = [
			[2,3], # these contexts are good match to query 1
			[1,3], # these contexts are good match to query 2
			[1,2], # these contexts are good match to query 3
			[5,6], # these contexts are good match to query 4
			[4,6], # these contexts are good match to query 5
			[4,5], # these contexts are good match to query 6
			[8,9], # these contexts are good match to query 7
			[7,9], # these contexts are good match to query 8
			[7,8], # these contexts are good match to query 9
			[11,12], # these contexts are good match to query 10
			[10,12], # these contexts are good match to query 11
			[10,11]  # these contexts are good match to query 12
		]

		for i in range(1, 3*4+1):
			top3 = sorted(
				enumerate(dots[i]), key=lambda x: x[1], reverse=True
			)[:3]
			top3_positions = [t[0] for t in top3]

			# both of the expected top matches should appear in the 
			# top 3 (the UNK token might also be a good match, which
			# is why we need to check the top 3, not just top 2).
			self.assertTrue(expected_tops[i-1][0] in top3_positions)
			self.assertTrue(expected_tops[i-1][1] in top3_positions)




class TestWord2Vec(TestCase):

	'''
	This tests comnponent Word2Vec functionality by supplying
	synthetic numerical data into its components, to make sure that
	the solutions are mathematically correct.  It doesn't test iteration
	over an actual text corpus, which is tested by TestWord2VecOnCorpus.
	'''

	def setUp(self):

		self.VOCAB_SIZE = 3
		self.NUM_EMBEDDING_DIMENSIONS = 4

		# Make a query input vector.  This holds indices that represent
		# words in the vocabulary.  For the test we have just three words
		# in the vocabulary
		self.TEST_INPUT = np.array(
			[[i / self.VOCAB_SIZE, i % self.VOCAB_SIZE] for i in range(9)]
		).astype('int32')

		# Artificially adopt this word embedding matrix for query words
		self.QUERY_EMBEDDING = np.array([
			[-0.04576914, -0.26519672, -0.06857708, -0.23748968],
			[ 0.08540803,  0.32099229, -0.19136694, -0.48263541],
			[-0.33319689,  0.26062664,  0.06826347, -0.39083191]
		])

		# Artificially adopt this word embedding matrix for context words
		self.CONTEXT_EMBEDDING = np.array([
			[-0.29474795, -0.2559814 , -0.04503929,  0.35159791],
			[ 0.00963128,  0.22368461,  0.44933862,  0.48584304],
			[ 0.05338832, -0.22895403, -0.08288041, -0.47226618],
		])



	def test_save_load_embedding(self):
		'''
		Test that save and load actually write and read the 
		Word2VecEmbedder's parameters, by testing that a saved and then 
		then loaded model yields the same model as the original.
		'''

		# Remove any saved file that may be left over from a previous run
		save_dir = 'test-data/test-w2v-embedder'
		if os.path.exists(save_dir):
			shutil.rmtree(save_dir)

		# Make a Word2VecEmbedder with pre-specified query and context
		# embeddings
		mock_input = shared(np.array([[]], dtype='int32'))
		embedder = Word2VecEmbedder(
			mock_input,
			batch_size=len(self.TEST_INPUT),
			vocabulary_size=self.VOCAB_SIZE,
			num_embedding_dimensions=self.NUM_EMBEDDING_DIMENSIONS,
			query_embedding_init=self.QUERY_EMBEDDING,
			context_embedding_init=self.CONTEXT_EMBEDDING
		)
		embedder.save(save_dir)

		# Create a new embedder, and attempt to load from file
		embedder = Word2VecEmbedder(
			mock_input,
			batch_size=len(self.TEST_INPUT),
			vocabulary_size=self.VOCAB_SIZE,
			num_embedding_dimensions=self.NUM_EMBEDDING_DIMENSIONS,
		)
		embedder.load(save_dir)

		W, C = embedder.get_params()

		self.assertTrue(np.allclose(self.QUERY_EMBEDDING, W.get_value()))
		self.assertTrue(np.allclose(self.CONTEXT_EMBEDDING, C.get_value()))

		# Remove the saved file
		shutil.rmtree(save_dir)



	def test_noise_contrastive_learning(self):
		'''
		Given a simple synthetic dataset, test that the Word2vecEmbedder,
		coupled with a loss function from get_noise_contrastive_loss, 
		produces a trainable system that learns the theoretically ideal 
		embeddings as expected from [1].  

		[1] "Noise-contrastive estimation of unnormalized statistical 
			models, with applications to natural image statistics"
			by Michael U Gutmann, and Aapo Hyvarinen
		'''

		# Seed randomness to make the test reproducible
		np.random.seed(1)

		# Make the positive input.  First component of each example is
		# the query input, and second component is the context.  In the
		# final embeddings that are learned, dotting these rows and columns
		# respectively from the query and context embedding matrices should
		# give higher values than any other row-column dot products.
		signal_examples = [
			[0,2], [1,3], [2,0], [3,1], [4,6], [5,7], [6,4], [7,5], [8,9], 
			[9,8]
		]

		# Predifine the size of batches and the embedding
		num_signal_examples = len(signal_examples)
		num_noise_examples = 100
		batch_size = num_signal_examples + num_noise_examples
		vocab_size = 10
		num_embedding_dimensions = 5

		# Mock a symbolic minibatch 
		# (it's empty but will be filled in a moment)
		symbolic_batch = shared(np.array([[]], dtype='int32'))

		# Make a Word2VecEmbedder object, feed it the mocked batch
		word2vec_embedder = Word2VecEmbedder(
			input_var=symbolic_batch,
			batch_size=batch_size,
			vocabulary_size=vocab_size,
			num_embedding_dimensions=num_embedding_dimensions
		)

		# Get the params and output from the word2vec embedder
		symbolic_output = word2vec_embedder.get_output()
		params = word2vec_embedder.get_params()

		# Define the loss function, and get parameter updates based on it
		loss = get_noise_contrastive_loss(
			symbolic_output, num_signal_examples)
		updates = nesterov_momentum(
			loss, params, learning_rate=0.1, momentum=0.9)

		# Create the training function.  No inputs because the "inputs" are
		# implemented as theano shared variables
		train = function([], loss, updates=updates)

		num_replicates = 5
		num_epochs = 3000
		embedding_products = []
		W, C = word2vec_embedder.get_params()
		start = time.time()
		for rep in range(num_replicates):
			W.set_value(np.random.normal(
				0, 0.01, (vocab_size, num_embedding_dimensions)
			).astype(dtype='float32'))
			C.set_value(np.random.normal(
				0, 0.01, (vocab_size, num_embedding_dimensions)
			).astype('float32'))
			for epoch in range(num_epochs):

				# Sample new noise examples every epoch
				noise_examples = [
					[i / 10, np.random.randint(0,10)]
					for i in range(num_noise_examples)
				]

				# Assemble and load the batch to the symbolic batch variable
				batch_data = np.array(
					signal_examples + noise_examples, dtype='int32'
				)
				symbolic_batch.set_value(batch_data)

				this_loss = train()

			embedding_product = np.dot(W.get_value(), C.get_value().T)
			embedding_products.append(usigma(embedding_product))

		mean_embedding_products = np.mean(embedding_products, axis=0)

		# We expect that the embeddings will allocate the most probability
		# to the contexts that were provided for words in the toy data.
		# We always provided a single context via batch_contexts
		# (e.g. context 2 provided for word 0), so we expect these contexts
		# to be the maximum.
		expected_max_prob_contexts = np.array(signal_examples)[:,1]
		self.assertTrue(np.array_equal(
			np.argmax(mean_embedding_products, axis=1),
			expected_max_prob_contexts
		))

		# The dot product of a given word embedding and context embedding
		# have an interpretation as the probability that that word and
		# context derived from the toy data instead of the noise.
		# See equation 3 in Noise-Contrastive Estimation of Unnormalized
		# Statistical Models, with Applications to Natural Image
		# StatisticsJournal of Machine Learning Research 13 (2012),
		# pp.307-361.
		# That shows the probability should be around 0.5
		# Since the actual values are stocastic, we check that the
		# average of repeated trials is within 0.25 - 0.75.
		embedding_maxima = np.max(mean_embedding_products, axis=1)
		self.assertTrue(all(
			[x > 0.25 for x in embedding_maxima]
		))
		self.assertTrue(all(
			[x < 0.75 for x in embedding_maxima]
		))


	def test_Word2VecEmbedder(self):
		'''
		Test that the architecture defined by Word2VecEmbedder generates the
		expected theano expressions for embedded inputs and for final output
		'''

		input_var = T.imatrix('input_var')

		embedder = Word2VecEmbedder(
			input_var,
			batch_size=len(self.TEST_INPUT),
			vocabulary_size=self.VOCAB_SIZE,
			num_embedding_dimensions=self.NUM_EMBEDDING_DIMENSIONS,
			query_embedding_init=self.QUERY_EMBEDDING,
			context_embedding_init=self.CONTEXT_EMBEDDING
		)

		query_embedding = embedder.query_embedding
		context_embedding = embedder.context_embedding
		dots = embedder.get_output()

		f = function([input_var], query_embedding)
		g = function([input_var], context_embedding)
		h = function([input_var], dots)

		# Calculate the embeddings and the output
		query_embeddings = f(self.TEST_INPUT)
		context_embeddings = g(self.TEST_INPUT)
		test_output = h(self.TEST_INPUT)

		# Calculate the expected embeddings and output
		expected_query_embeddings = np.repeat(
			self.QUERY_EMBEDDING, 3, axis=0
		)
		expected_context_embeddings = np.tile(
			self.CONTEXT_EMBEDDING, (3,1)
		)
		expected_output = usigma(np.dot(
			expected_query_embeddings, expected_context_embeddings.T
		)).diagonal()

		# Check for equality between all found and expected values
		self.assertTrue(np.allclose(
			query_embeddings, expected_query_embeddings
		))
		self.assertTrue(np.allclose(
			context_embeddings, expected_context_embeddings
		))
		self.assertTrue(np.allclose(test_output, expected_output))



if __name__=='__main__':
	main()



'''
This module keeps track of the vocabulary present in a corpus, provides
a two-way maping between tokens (strings) and token_ids (integers),
keeps track of token frequencies, and can yield samples from the
unigram distribution.
'''

import os
from token_map import TokenMap, SILENT, WARN, ERROR, UNK
from counter_sampler import CounterSampler

class UnigramDictionary(object):
	'''
	Bundles together a TokenMap and CounterSampler.  Provides a method for
	pruning the vocabluary while keeping the TokenMap and CounterSampler
	in sync with one another.
	'''


	def __init__(self, on_unk=WARN, token_map=None, counter_sampler=None):
		'''
		Create a new UnigramDictionary.  Typical usage provides no
		arguments, but a token_map and counter_sampler can be provided
		to build a UnigramDictionary that comprises them.
		'''
		self.on_unk = on_unk
		self.token_map = token_map
		if token_map is None:
			self.token_map = TokenMap(on_unk=on_unk)

		self.counter_sampler = counter_sampler
		if counter_sampler is None:
			self.counter_sampler = CounterSampler()


	def sort(self):
		unk_count = self.counter_sampler.counts[0]

		# Get the counts and tokens (skipping the first UNK entry)
		# They are parallel arrays (ith count corresponds to ith token)
		counts = self.counter_sampler.counts[1:]
		tokens = self.token_map.tokens[1:]

		# Zip them together and sort by counts
		token_counts = zip(counts, tokens)
		token_counts.sort(reverse=True)

		# Separate them again
		new_counts = [unk_count]
		new_tokens = ['UNK']
		for count, token in token_counts:
			new_counts.append(count)
			new_tokens.append(token)

		# Rebuild the token_map and counter_sampler on the sorted arrays
		self.token_map = TokenMap(on_unk=self.on_unk, tokens=new_tokens)
		self.counter_sampler = CounterSampler(counts=new_counts)


	def remove(self, token):
		idx = self.get_id(token)
		self.token_map.remove(token)
		self.counter_sampler.remove(idx)


	def compact(self):
		self.token_map.compact()
		self.counter_sampler.compact()


	def prune(self, min_frequency=5):
		'''
		Remove all tokens that have been observed fewer than min_frequency
		times.  Counts for tokens that are removed are attributed to UNK.
		'''
		counts = []
		tokens = []
		for idx, token in enumerate(self.token_map.tokens):

			# Copy over tokens that have at least min_frequency
			# observations. Also copy over UNK no matter what it's
			# frequency.
			if (
				self.counter_sampler.get_frequency(idx) >= min_frequency
				or idx == 0
			):
				tokens.append(token)
				counts.append(self.get_frequency(idx))

			# Skip tokens that have too little frequency.  Attribute their
			# observations to UNK
			else:
				counts[UNK] += self.get_frequency(idx)

		# Create a new TokenMap and CounterFrequency based on the
		# filtered tokens and their counts
		self.token_map = TokenMap(on_unk=self.on_unk, tokens=tokens)
		self.counter_sampler = CounterSampler(counts=counts)


	def add(self, token):
		'''
		Add a new token.  If this "token type" (which means this specific
		spelling of a word) has not been seen before, add it to the
		mapping.  Also increment the count for that token type.  Return
		its ID under the token mapping.
		'''

		# Get or create an id for this token
		token_id = self.token_map.add(token)

		# Increment the frequency count
		self.counter_sampler.add(token_id)

		return token_id


	def get_vocab_size(self):
		'''
		Return the number of unique tokens in the token_map.
		'''
		return len(self.token_map)


	def get_num_tokens(self):
		'''
		Return the total number of (non-distinct) tokens observed.
		'''
		return len(self.counter_sampler)


	def __len__(self):
		'''
		Same as get_vocab_size().
		Return the number of unique tokens in the token_map.
		'''
		return len(self.token_map)


	def update(self, token_iterable):
		return [self.add(token) for token in token_iterable]


	def get_id(self, token):
		'''
		Get the id (int) for the corresponding token (string).
		'''
		# Delegate to the underlying token_map.
		return self.token_map.get_id(token)


	def get_ids(self, token_iterable):
		'''
		Get the ids (list of ints) for the corresponding tokens (strings)
		issued by token_iterable.
		'''
		# Delegate to the underlying token map.
		return self.token_map.get_ids(token_iterable)


	def get_token(self, idx):
		'''
		Return token (string) for the corresponding id (int)
		'''
		# Delegate to the underlying token map
		return self.token_map.get_token(idx)


	def get_tokens(self, idx_iterable):
		'''
		Return the tokens (list of strings) for the corresponding ids
		(ints) issued by idx_iterable.
		'''
		# Delegate to the underlying token map.
		return self.token_map.get_tokens(idx_iterable)


	def save(self, savedir):
		'''
		Save the UnigramDictionary to the directory specified.  This saves
		the underlying TokenMap and CounterSampler in the directory
		given (savedir), using the default filenames "token-map.gz" and
		"counter-sampler.gz".
		'''

		# If the directory provided is a file, raise an error
		if os.path.exists(savedir):
			if os.path.isfile(savedir):
				raise IOError(
					'Directory specified for saving UnigramDictionary is a '
					'file.'
				)

		# If the directory provided doesn't exist, make it (this will not
		# make parent directories though).
		else:
			os.mkdir(savedir)


		# Save the TokenMap and CounterSampler by delegating to their
		# save functions.
		self.token_map.save(os.path.join(savedir, 'token-map.gz'))
		self.counter_sampler.save(os.path.join(
			savedir, 'counter-sampler.gz'
		))


	def load(self, loaddir):
		'''
		Load a UnigramDictionary from the specified directory, by
		loading the TokenMap and CounterSampler stored there.  This assumes
		the filenames are 'token-map.gz' and 'counter-sampler.gz'.
		'''
		# Load the TokenMap by delegation to its load function
		self.token_map = TokenMap()
		self.token_map.load(os.path.join(loaddir, 'token-map.gz'))

		# Load the CounterSampler by delegation to its load function
		self.counter_sampler = CounterSampler()
		self.counter_sampler.load(
			os.path.join(loaddir, 'counter-sampler.gz'))


	def sample(self, shape=None):
		'''
		Draw a sample according to the counter_sampler probability
		'''
		# Delegate to the underlying CounterSampler
		return self.counter_sampler.sample(shape)


	def get_probability(self, token_id):
		'''
		Return the probability associated to token_id.
		'''
		# Delegate to the underlying CounterSampler
		return self.counter_sampler.get_probability(token_id)


	def get_frequency(self, token_id):
		'''
		Return the frequency associated to token_id.
		'''
		# Delegate to the underlying CounterSampler
		return self.counter_sampler.get_frequency(token_id)



import os

# Only import theano and lasagne if environment permits it
exclude_theano_set = 'EXCLUDE_THEANO' in os.environ
if exclude_theano_set and int(os.environ['EXCLUDE_THEANO']) == 1:
	# Don't import theano and lasagne
	pass
else:
	# Do import theano and lasagne
	import theano.tensor as T
	from theano import function
	import lasagne


def noise_contrast(signal, noise, scale=True):
	'''
	Takes the theano symbolic variables `signal` and `noise`, whose
	elements are interpreted as probabilities, and creates a loss
	function which rewards large probabilities in signal and penalizes
	large probabilities in noise.

	`signal` and `noise` can have any shape.  Their contributions will be
	summed over all dimensions.

	If `scale` is true, then scale the loss function by the size of the
	signal tensor --- i.e. divide by the signal batch size.  This makes
	the scale of the loss function invariant to changes in batch size
	'''

	signal_score = T.log(signal).sum()
	noise_score = T.log(1-noise).sum()
	objective = signal_score + noise_score
	loss = -objective

	loss = loss / signal.shape[0]

	return loss


def get_noise_contrastive_loss(activation, num_signal, scale=True):
	'''
	Convenience function to get the noise_contrast expression by specifying a
	single batch of outputs and giving the number of entries along axis 0,
	corresponding to signal activations (all others are assumed to be noise)
	activations.

	Differs from noise_contrast only in that noise contrast expects two
	theano variables, one for signal activations and one for noise.  This means
	the caller would have to separate signal from noise first.
	'''
	signal_activation = activation[0:num_signal,]
	noise_activation = activation[num_signal:,]
	return noise_contrast(signal_activation, noise_activation, scale)



from noise_contrast import get_noise_contrastive_loss
import os
import numpy as np
from dataset_reader import DatasetReader, default_parse
from theano_minibatcher import (
	TheanoMinibatcher, NoiseContrastiveTheanoMinibatcher
)

import os
# Only import theano and lasagne if environment permits it
exclude_theano_set = 'EXCLUDE_THEANO' in os.environ
if exclude_theano_set and int(os.environ['EXCLUDE_THEANO']) == 1:
	# Don't import theano and lasagne
	pass
else:
	from theano import tensor as T, function, shared
	import lasagne
	from lasagne.layers import (
		get_output, InputLayer, EmbeddingLayer, get_all_params,
		get_all_param_values
	)
	from lasagne.init import Normal
	from lasagne.updates import nesterov_momentum


def row_dot(matrixA, matrixB):
	C = (matrixA * matrixB).sum(axis=1)
	return C


def sigmoid(tensor_var):
	return 1/(1+T.exp(-tensor_var))


def word2vec(

		# Input / output options
		files=[],
		directories=[],
		skip=[],
		save_dir=None,
		read_data_async=True,
		num_processes=3,
		max_queue_size=0,
		parse=default_parse,

		# Batching options
		num_epochs=5,
		batch_size = 1000,  # Number of *signal* examples per batch
		macrobatch_size = 100000,

		# Dictionary options
		unigram_dictionary=None,
		load_dictionary_dir=None,
		min_frequency=10,

		# Sampling options
		noise_ratio=15,
		kernel=[1,2,3,4,5,5,4,3,2,1],
		t = 1.0e-5,

		# Embedding options
		num_embedding_dimensions=500,
		query_embedding_init=None,
		context_embedding_init=None,

		# Learning rate options
		learning_rate=0.1,
		momentum=0.9,

		# Verbosity option
		verbose=True
	):

	'''
	Helper function that handles all concerns involved in training
	A word2vec model using the approach of Mikolov et al.  It surfaces
	all of the options.

	For customizations going beyond simply tweeking existing options and
	hyperparameters, substitute this function by writing your own training
	routine using the provided classes.  This function would be a starting
	point for you.
	'''

	# Make a Word2VecMinibatcher, pass through parameters sent by caller
	reader = DatasetReader(
		files=files,
		directories=directories,
		skip=skip,
		macrobatch_size=macrobatch_size,
		max_queue_size=max_queue_size,
		noise_ratio=noise_ratio,
		num_processes=num_processes,
		unigram_dictionary=unigram_dictionary,
		load_dictionary_dir=load_dictionary_dir,
		min_frequency=min_frequency,
		t=t,
		kernel=kernel,
		parse=parse,
		verbose=verbose
	)

	# Prepare the dataset reader (this produces the counter_sampler stats)
	if not reader.is_prepared():
		if verbose:
			print 'preparing dictionaries...'
		reader.prepare(save_dir=save_dir)

	# Make a symbolic minibatcher
	minibatcher = NoiseContrastiveTheanoMinibatcher(
		batch_size=batch_size,
		noise_ratio=noise_ratio,
		dtype="int32",
		num_dims=2
	)

	# Make a Word2VecEmbedder object, feed it the combined input.
	# Note that the full batch includes noise examples and signal_examples
	# so is larger than batch_size, which is the number of signal_examples
	# only per batch.
	full_batch_size = batch_size * (1 + noise_ratio)
	embedder = Word2VecEmbedder(
		input_var=minibatcher.get_batch(),
		batch_size=full_batch_size,
		vocabulary_size=reader.get_vocab_size(),
		num_embedding_dimensions=num_embedding_dimensions,
		query_embedding_init=query_embedding_init,
		context_embedding_init=context_embedding_init
	)

	# Architectue is ready.  Make the loss function, and use it to create 
	# the parameter updates responsible for learning
	loss = get_noise_contrastive_loss(embedder.get_output(), batch_size)
	updates = nesterov_momentum(
		loss, embedder.get_params(), learning_rate, momentum
	)

	# Include minibatcher updates, which cause the symbolic batch to move
	# through the dataset like a sliding window
	updates.update(minibatcher.get_updates())

	# Use the loss function and the updates to compile a training function.
	# Note that it takes no inputs because the dataset is fully loaded using
	# theano shared variables
	train = function([], loss, updates=updates)

	# Iterate through the dataset, training the embeddings
	for epoch in range(num_epochs):

		if verbose:
			print 'starting epoch %d' % epoch

		if read_data_async:
			macrobatches = reader.generate_dataset_parallel()
		else:
			macrobatches = reader.generate_dataset_serial()

		macrobatch_num = 0
		for signal_macrobatch, noise_macrobatch in macrobatches:

			macrobatch_num += 1
			if verbose:
				print 'running macrobatch %d' % (macrobatch_num - 1)

			minibatcher.load_dataset(signal_macrobatch, noise_macrobatch)
			losses = []
			for batch_num in range(minibatcher.get_num_batches()):
				if verbose:
					print 'running minibatch', batch_num
				losses.append(train())
			if verbose:
				print '\taverage loss: %f' % np.mean(losses)

	# Save the model (the embeddings) if save_dir was provided
	if save_dir is not None:
		embedder.save(save_dir)
		reader.save_dictionary(save_dir)

	# Return the trained embedder and the dictionary mapping tokens
	# to ids
	return embedder, reader



class Word2VecEmbedder(object):

	def __init__(
		self,
		input_var,
		batch_size,
		vocabulary_size=100000,
		query_vocabulary_size=None,
		context_vocabulary_size=None,
		num_embedding_dimensions=500,
		query_embedding_init=None,
		context_embedding_init=None
	):

		self.query_embedding_init = query_embedding_init
		if query_embedding_init is None:
			self.query_embedding_init = Normal()

		self.context_embedding_init = context_embedding_init
		if context_embedding_init is None:
			self.context_embedding_init = Normal()

		# If only vocabular_size is specified, then both 
		# query_vocabulary_size and context_vocabulary_size take on that 
		# value, but ppecific values given for query_ or 
		# context_vocabulary_size override.
		self.query_vocabulary_size = vocabulary_size
		if query_vocabulary_size is not None:
			self.query_vocabulary_size = query_vocabulary_size

		self.context_vocabulary_size = vocabulary_size
		if context_vocabulary_size is not None:
			self.context_vocabulary_size = context_vocabulary_size

		self.input_var = input_var
		self.batch_size = batch_size
		self.num_embedding_dimensions = num_embedding_dimensions
		self._embed = None

		# Every row (example) in a batch consists of a query word and a
		# context word.  We need to learn separate embeddings for each,
		# and so, the input will be separated into all queries and all
		# contexts, and handled separately.
		self.query_input = input_var[:,0]
		self.context_input = input_var[:,1]

		# Make separate input layers for query and context words
		self.l_in_query = lasagne.layers.InputLayer(
			shape=(self.batch_size,), input_var=self.query_input
		)
		self.l_in_context = lasagne.layers.InputLayer(
			shape=(self.batch_size,), input_var=self.context_input
		)

		# Make separate embedding layers for query and context words
		self.l_embed_query = lasagne.layers.EmbeddingLayer(
			incoming=self.l_in_query,
			input_size=self.query_vocabulary_size,
			output_size=self.num_embedding_dimensions,
			W=self.query_embedding_init
		)
		self.l_embed_context = lasagne.layers.EmbeddingLayer(
			incoming=self.l_in_context,
			input_size=self.context_vocabulary_size,
			output_size=self.num_embedding_dimensions,
			W=self.context_embedding_init
		)

		# Obtain the embedded query words and context words
		self.query_embedding = get_output(self.l_embed_query)
		self.context_embedding = get_output(self.l_embed_context)

		# We now combine the query and context embeddings, taking the
		# dot product between corresponding queries and contexts.
		# We can express this as an element-wise multiplication between
		# the array of all query embeddings with all context embeddings
		# Summing along the rows of the resulting matrix yields the dot
		# product for each query-context pair.
		self.match_scores = row_dot(
			self.query_embedding,
			self.context_embedding
		)
		#self.match_scores = T.dot(
		#	self.query_embedding, self.context_embedding.T
		#).diagonal()

		# Finally apply the sigmoid activation function
		self.output = sigmoid(self.match_scores)


	def embed(self, token_id):
		'''
		Return the vectorspace embedding for the token.

		INPUTS
		* token [int] or [iterable : int]: Integer representation of a
			single token (e.g. a word), or a list of tokens.
		'''
		if self._embed is None:
			self._compile_embed()

		# This function is overloaded to accept a single token_id or
		# a list thereof.  Resolve that here.
		if isinstance(token_id, int):
			token_ids = [token_id]
		else:
			token_ids = token_id

		return self._embed(token_ids)


	def _compile_embed(self):
		'''
		Compiles the embedding function.  This is a separate theano
		function from that used during training of the network, but it
		produces exactly the same embeddings.  After being compiled
		once (usually the first time self.embed() is called), it will
		remain up-to-date, supplying the correct embeddings even if
		more training occurs.  This is because it references the same
		theano shared variables for its embedding parameters as the
		embedding layer used in training.

		(No Inputs)
		OUTPUTS
		* [None]
		'''
		input_tokens = T.ivector('token')
		l_in = lasagne.layers.InputLayer(
			shape=(None,), input_var=input_tokens
		)
		l_emb = lasagne.layers.EmbeddingLayer(
			incoming=l_in,
			input_size=self.query_vocabulary_size,
			output_size=self.num_embedding_dimensions,
			W=self.l_embed_query.W
		)
		embedding = get_output(l_emb)
		self._embed = function([input_tokens], embedding)


	def get_param_values(self):
		'''
		returns a list of the trainable parameters *values*, as
		np.ndarray's.
		'''
		return (
			get_all_param_values(self.l_embed_query, trainable=True) +
			get_all_param_values(self.l_embed_context, trainable=True)
		)


	def get_params(self):
		'''returns a list of the trainable parameters, that is, the query
		and context embeddings.  (similar to layer.get_all_params.)'''
		return (
			self.l_embed_query.get_params(trainable=True) +
			self.l_embed_context.get_params(trainable=True)
		)


	def get_output(self):
		'''returns the symbolic output.  (similar to layer.get_output.)'''
		return self.output


	def save(self, directory):
		'''
		Saves the model parameters (embeddings) to disk, in a file called
		"embeddings.npz" under the directory given.
		'''

		# We are willing to create the directory given if it doesn't exist
		if not os.path.exists(directory):
			os.mkdir(directory)

		# Save under the directory given in a file called "embeddings.npz'
		save_path = os.path.join(directory, "embeddings.npz")

		# Get the parameters and save them to disk
		W, C = self.get_param_values()
		np.savez(save_path, W=W, C=C)


	def load(self, directory):
		'''
		Loads the model parameter values (embeddings) stored in the
		directory given.  Expects to find the parameters in a file called
		"embeddings.npz" within the directory given.
		'''

		# By default, we save to a file called "embeddings.npz" within the
		# directory given to the save function.
		save_path = os.path.join(directory, "embeddings.npz")

		# Load the parameters
		npfile = np.load(save_path)
		W_loaded = npfile['W'].astype('float32')
		C_loaded = npfile['C'].astype('float32')

		# Get the theano shared variables that are used to hold the
		# parameters, and fill them with the values just loaded
		W_shared, C_shared = self.get_params()
		W_shared.set_value(W_loaded)
		C_shared.set_value(C_loaded)



from iterable_queue import IterableQueue
from multiprocessing import Process
from categorical import Categorical
from token_map import UNK
from unigram_dictionary import UnigramDictionary
import numpy as np
from theano import shared
import os
import sys


def default_parse(filename):
	tokenized_sentences = []
	for line in open(filename):
		tokenized_sentences.append(line.strip().split())
	return tokenized_sentences


MAX_NUMPY_SEED = 4294967295
def reseed():
	'''
	Makes a hop in the random chain.

	If called before spawning a child processes, it will ensure each child
	generates different random numbers.  Unlike seeding child randomness
	from an os source of randomness, this will produce globally consistent
	results when the parent starts with the same random seed.

	In other words, child processes will all have different random chains
	but the *particular* random chains they have is deterministic and
	reproducible when the parent process randomness is seeded to the same
	value.  So child processes have samples that are independent from one
	another, but reproducible by controlling the seed of the parent process.
	'''
	np.random.seed(np.random.randint(MAX_NUMPY_SEED))


class Minibatcher(object):

	def __init__(
		self,
		files=[],
		directories=[],
		skip=[],
		batch_size = 1000,
		num_example_generators=3,
		verbose=True
	):

		# Register parameters to instance namespace
		self.files = files
		self.directories = directories
		self.skip = skip
		self.batch_size = batch_size
		self.num_example_generators = num_example_generators
		self.verbose = verbose

		self.setup_symbolic_minibatching()

	def check_access(self, savedir):

		savedir = os.path.abspath(savedir)
		path, dirname = os.path.split(savedir)

		# Make sure that the directory we want exists (make it if not)
		if not os.path.isdir(path):
			raise IOError('%s is not a directory or does not exist' % path)
		if not os.path.exists(savedir):
			os.mkdir(savedir)
		elif os.path.isfile(savedir):
			raise IOError('%s is a file. % savedir')

		# Make sure we can write to the file
		f = open(os.path.join(
			savedir, '.__test-minibatch-generator-access'
		), 'w')
		f.write('test')
		f.close
		os.remove(os.path.join(
			savedir, '.__test-minibatch-generator-access'
		))


	def preparation(self):
		'''
		Used to perform any preparation steps that are needed before
		minibatching can be done (not always necesary).  E.g. assembling a
		dictionary that maps tokens to integers, and determining the total
		vocabulary size of the corpus. Or normalizing variables based on
		their total variation throughout the corpus.

		Usually not called directly, but rather, `prepare()` is called,
		and delegates to this method.  Prepare also enables saving data
		generated during this step.  To do so, override the `save()` method
		(which is a no-op by default).
		'''
		pass


	def save(self, dirname):
		'''
		This can be overridden and used to save data generated within
		`preparation()`.  By default this is a no-op.

		INPUTS
		* directory [str]: path to a directory in which to place files
			to be saved.  Will be created if does not exist (containing
			directory must exist).
		'''
		pass


	def prepare(self, savedir=None, *args, **kwargs):
		'''
		Used to perform any preparation steps that are needed before
		minibatching can be done.  E.g. assembling a dictionary that
		maps tokens to integers, and determining the total vocabulary size
		of the corpus.  It is assumed that files will need
		to be saved as part of this process, and that they should be
		saved under `savedir`, with `self.save()` managing the details
		of writing files under `savedir`.

		INPUTS

		* Note About Inputs *
		the call signature of this method is variable and is
		determined by the call signature of the core
		`self.preparation()` method.  Refer to that method's call
		signature.  Minimally, this method accepts `savedir`

		* savedir [str]: path to directory in which preparation files
			should be saved.

		RETURNS
		* [None]
		'''

		# Before doing anything, if we were requested to save the
		# dictionary, make sure we'll be able to do that (fail fast)
		if savedir is not None:
			self.check_access(savedir)

		self.preparation(savedir, *args, **kwargs)

		# Save the dictionary, if requested to do so.
		if savedir is not None:
			self.save(savedir)


	def generate_filenames(self):

		'''
		Generator that yields all filenames (absolute paths) making up the
		dataset.  (Files are specified to the Minibatcher constructor
		files and / or directories.  All listed files and all files directly
		contained in listed directories will be processed, unless they
		match regex patterns in the optional `skip` list.

		(no INPUTS)

		YIELDS
		* [str]: absolute path to a dataset file
		'''

		# Process all the files listed in files, unles they match an
		# entry in skip
		#print 'starting reading'
		if self.files is not None:
			for filename in self.files:
				filename = os.path.abspath(filename)

				# Skip files if they match a regex in skip
				if any([s.search(filename) for s in self.skip]):
					continue

				if self.verbose:
					print '\tprocessing', filename

				yield filename

		# Process all the files listed in each directory, unless they
		# match an entry in skip
		if self.directories is not None:
			for dirname in self.directories:
				dirname = os.path.abspath(dirname)

				# Skip directories if they match a regex in skip
				if any([s.search(dirname) for s in self.skip]):
					continue

				for filename in os.listdir(dirname):
					filename = os.path.join(dirname, filename)

					# Only process the *files* under the given directories
					if not os.path.isfile(filename):
						continue

					# Skip files if they match a regex in skip
					if any([s.search(filename) for s in self.skip]):
						continue

					if self.verbose:
						print '\tprocessing', filename

					yield filename


	def parse(self, filename):
		'''
		Parses input corpus files into a file-format-independent in-memory
		representation.  Since training examples are read from file,
		this function provides file-reading and parsing functionality,
		and should bring the data stored in corpus files into a form that
		makes generating actual training examples easy.  This function,
		however, should not actually generate training examples; this
		should be done in `build_examples()`.  Rather, this provides a
		modular component that can be exchaged in cases where the file
		format changes, making the example-building logic independent of
		file format logic

		INPUTS
		* filename [str]: path to a corpus file to be read

		RETURNS
		* [any]: file-format-independent representation of training data.
		'''

		# This example parser simply returns a list of all the lines in
		# the file in which each line is tokenized.  Typically this
		# function will need to do more, such as loading json, numpy, or
		# image data file formats.  The return value will be provided to
		# the method `build_examples`.
		return default_parse(filename)


	def build_examples(self, parsed):
		'''
		Builds and yields training examples based on parsed data obtained
		as output from `parse()`.  This should contain all of the logic
		for building training examples based on data from dataset files,
		but should be independent of the dataset storage format (parsing
		the storage format should be handled in `parse()`.

		INPUTS
		* parsed [any]: Data from a single dataset file that has been
			parsed into a file-format-independent form.

		YIELDS
		* [any]: Usually a single training example.  However, depending
			on how you implement `batch_examples()`, this may not be
			necessary.
		'''
		# This minimalistic example assumes that parsed has yielded an
		# iterable of items, and that each item is already a well-formed
		# training example, which can directly be yielded.  Normally
		# substantial transformations and sub-sampling might be done
		# here to generate training examples using the dataset.
		for item in parsed:
			example = item
			yield example


	def process_file(self, filename):
		'''
		Generator that encapsulates all processing for a single data file,
		i.e., parsing and building of examples.

		INPUTS
		* filename [str]: path to file to be processed.

		YIELDS
		* example [any]: object representing (usually) one training
			example.
		'''

		self.cur_file = filename
		parsed = self.parse(filename)
		examples = self.build_examples(parsed)
		for example in examples:
			yield example


	def generate_examples(self):
		'''
		Generates training examples.  Via subroutienes, this iterates
		over all data files, parsing them, and then generating
		training examples from them, which are yielded one at a time.

		(no INPUTS)

		YIELDS
		* [any]: (Usually) single training example.
		'''

		for filename in self.generate_filenames():
			for example in self.process_file(filename):
				yield example


	def generate_examples_async(self, file_queue, example_queue):
		'''
		Asynchronous version of `generate_examples()` designed to be
		run within a forked worker process.  Draws filenames of dataset
		files from a queue, and yields training examples onto a queue.
		This is part of the minibatching pipeline used for asynchronous
		minibatching, which is invoked when Minibatcher is treated
		as an iterable.

		INPUTS
		* file_queue [IterableQueue.ConsumerQueue]: A queue containing
			absolute paths ([str]s) of dataset files.
		* example_queue [IterableQueue.ProducerQueue]: A queue that accepts
			training examples ([any]).

		(no OUTPUS)

		CONSUMES
		* filename [str]: Absolute path to a dataset file.

		PRODUCES
		* example [any]: A single training example
		'''
		for filename in file_queue:
			for example in self.process_file(filename):
				example_queue.put(example)

		example_queue.close()


	def batch_examples(self, example_iterator):
		'''
		Accumulates examples from an example_iterator bundling them into
		minibatches, which are yielded.

		INPUTS
		* example_iterator [iterable]: An iterable that yields examples
			([any]).

		YIEDLS
		* minibatches [iterable]: An iterable that yields examples ([any]).
		'''
		minibatch = []
		for example in example_iterator:
			batch.append(example)
			if len(minibatch) == self.batch_size:
				yield minibatch
				minibatch = []
		if len(minibatch) > 0:
			yield minibatch


	def generate_minibatches(self):
		'''
		Full-pass non-multiprocessing generator of minibatches.
		Iterates over all datafiles, parses them, generates training
		examples, and batches those training examples into minibatches,
		which are yielded.  Can be used as an alternative to `__iter__()`.
		Minibatches will not be prepared in the background (normally
		preparing in the background offers speed advantages).  Instead
		minibatches will be constructed on demand according to the
		pace of looping by the caller.  Normally much slower, but possibly
		desireable if you don't want background batching processes to be
		spawned.

		(no INPUTS)

		YIELDS
		* minibatches [iterable]: Iterable of training examples ([any]).
		'''
		for minibatch in self.batch_examples(self.generate_examples()):
			yield minibatch


	def get_minibatches(self):
		'''
		Full-pass non-multiprocessing generator for minibatches.  Differs
		from `generate_minibatches()` in that it assembles and returns
		an ordinary python list containing the entire training set.
		This can be useful if you want to load the full training set
		into memory for some reason.  However, this is generally slower
		than `__iter__()` because batching does not take place in the
		background using spawned processes, but rather takes place in
		the foreground, blocking until the entire training set has been
		read into memory.  This will usually have similar training speed
		characteristics to `generate_minibatches()`, but will block for
		a long time at the start while data is read into memory, and then
		will catch up as training takes place without the need for further
		reads from storage. Large datasets will take long to load and will
		consume large amounts of memory.

		(no INPUTS)

		OUTPUTS
		* [list : [iterable]]: A list of minibatches.
		'''
		minibatches = []
		for minibatch in self.generate_minibatches():
			minibatches.append(minibatch)

		return minibatches


	def generate_minibatches_async(self, example_queue, minibatch_queue):
		'''
		Asynchronous version of `generate_minibatches()`, designed to be
		run within a forked worker process.  Used in conjuction with
		`generate_examples_async()`, via the sharing of an example_queue,
		to create an asynchronous pipeline that assembles minibatches in
		the background.  This is the default method used for minibatching
		when Minibatcher is treated as an iterator.  See
		`__iter__()`.

		INPUTS
		* example_queue [IterableQueue.ConsumerQueue]: A queue that yields
			training examples ([any]).

		(no OUTPUS)

		CONSUMES
		* [any]: Training examples.

		PRODUCES
		* [iterable]: Minibatches of training examples.
		'''
		for minibatch in self.batch_examples(example_queue):
			minibatch_queue.put(minibatch)
		minibatch_queue.close()


	def get_async_batch_iterator(self):
		'''
		Builds an asynchronous minibatching pipeline, which reads all
		dataset files, parses them, generates training examples, and
		packages those training examples into minibatches.  Finally,
		it yields an iterable of minibatches, taking the form of an
		IterableQueue.ConsumerQueue.

		(no Inputs)

		OUTPUTS
		* [iterable (IterableQueue.ConsumerQueue)]: Iterable of
			minibatches.
		'''

		# TODO: currently the only randomness in minibatching comes from
		# the signal context and noise contexts that are drawn for a
		# given entity query tuple.  But the entity query tuples are read
		# deterministically in order through the corpus  Ideally examples
		# should be totally shuffled..


		file_queue = IterableQueue()
		example_queue = IterableQueue()
		minibatch_queue = IterableQueue()

		# Fill the file queue
		file_producer = file_queue.get_producer()
		for filename in self.generate_filenames():
			file_producer.put(filename)
		file_producer.close()

		# Make processes that process the files and put examples onto
		# the example queue
		for i in range(self.num_example_generators):

			# These calls to np.random are a hack to ensure that each
			# child example-generating process gets different randomness
			#reseed()
			Process(target=self.generate_examples_async, args=(
				file_queue.get_consumer(),
				example_queue.get_producer()
			)).start()

		# Make a processes that batches the files and puts examples onto
		# the minibatch queue
		Process(target=self.generate_minibatches_async, args=(
			example_queue.get_consumer(),
			minibatch_queue.get_producer()
		)).start()

		# Before closing the queues, make a consumer that will be used for
		# yielding minibatches to the external call for iteration.
		self.minibatch_consumer = minibatch_queue.get_consumer()

		# Close all queues
		file_queue.close()
		example_queue.close()
		minibatch_queue.close()

		# Return the minibatch_consumer as the iterator
		return self.minibatch_consumer


	def setup_symbolic_minibatching(self):

		# Create a shared variable that will (later) hold the entire dataset
		self.loaded_dataset = shared(np.array([],dtype='float32'))

		# Define a symbolic variable to represent a single minibatch,
		# based on a slice of the loaded dataset that is indexed by another
		# shared variable
		i = shared(0)
		self.X = self.loaded_dataset[i : i+batch_size,]

		# Define updates that causes the minibatch window to move
		self.updates = [(i, i+batch_size)]


	def get_symbolic_minibatch(self):
		# Return the symbolic minibatch and the updates
		return self.X, self.updates


	def get_symbolic_minibatch(self):
		'''
		This generates a theano shared variable storing the full dataset
		-- all training examples.  When the theano device setting is the
		GPU, shared variables are stored on the GPU, so this has the
		effect of loading the full dataset onto the GPU.

		One of the return values is a (set of) symbolic theano variable(s)
		corresponding to a single minibatch of the data.  This symbolic
		variable can be used to set up the training function.  What will
		happen during training is that this variable acts as a sliding
		"window" on the full dataset, selecting each minibatch in turn,
		even though the entire dataset is loaded into GPU memory.

		The indexing that causes the symbolic minibatch to address different
		parts of the dataset is itself a shared variable, and it can be
		updated using an update tuple provided to the updates list of a
		theano function.  The necessary update tuple is also provided as
		a return value, so that it can be incorporated into the training
		function
		'''

		# Load the entire dataset into RAM.
		# Use the async_batch_iterator to allow multiple read processes.
		dataset = []
		num_batches = 0
		for minibatch in self.get_async_batch_iterator():
			dataset.extend(minibatch)
			num_batches += 1
		dataset = np.array(dataset, dtype="float32")

		# Now move the dataset onto GPU (into GRAM).
		self.loaded_dataset.set_value(dataset)

		# Return the symbolic minibatch and the updates
		return num_batches


	def __iter__(self):
		return self.get_async_batch_iterator()


class NoiseSymbolicMinibatcherMixin(object):

	def setup_symbolic_minibatching(self):

		# Define a shared variable that will (later) hold entire dataset
		# Note that in this case we want it to hold integers, not floats
		# because the word2vec data are integers representing token ids.
		#
		# Here we are just creating an empty shared variable, but we
		# ensure that it has the right number of dimensions (2, because of
		# the [[]] passed into np.array), and correct dtype.
		self.loaded_dataset = shared(np.array([[]],dtype='int32'))

		# Define a the minibatch window indexing around a shared
		# variable, i, which tracks minibatch iteration
		i = shared(0)
		window_size = self.batch_size * (self.noise_ratio + 1)
		signal_start = i * window_size
		signal_end = signal_start + self.batch_size
		noise_start = signal_end
		noise_end = noise_start + self.batch_size * self.noise_ratio

		# Define a symbolic minibatch in terms of the full dataset
		# using the window indexing
		self.symbolic_signal_batch = self.loaded_dataset[
			signal_start : signal_end,]
		self.symbolic_noise_batch = self.loaded_dataset[
			noise_start : noise_end,]

		# Define an update that increments the minibatch
		self.updates = [(i, i+1)]


	def get_symbolic_minibatch(self):
		return (
			self.symbolic_signal_batch, self.symbolic_noise_batch,
			self.updates
		)


	def load_dataset(self):
		'''
		This generates a theano shared variable storing the full dataset
		-- all training examples.  When the theano device setting is the
		GPU, shared variables are stored on the GPU, so this has the
		effect of loading the full dataset onto the GPU.

		One of the return values is a (set of) symbolic theano variable(s)
		corresponding to a single minibatch of the data.  This symbolic
		variable can be used to set up the training function.  What will
		happen during training is that this variable acts as a sliding
		"window" on the full dataset, selecting each minibatch in turn,
		even though the entire dataset is loaded into GPU memory.

		The indexing that causes the symbolic minibatch to address different
		parts of the dataset is itself a shared variable, and it can be
		updated using an update tuple provided to the updates list of a
		theanod function.  The necessary update tuple is also provided as
		a return value, so that it can be incorporated into the training
		function
		'''

		# Load the entire dataset into RAM, converting to a numpy array.
		# Use the async_batch_iterator to allow multiple read processes.
		dataset = []
		num_batches = 0
		for signal_batch, noise_batch in self.get_async_batch_iterator():
			dataset.extend(signal_batch)
			dataset.extend(noise_batch)
			num_batches += 1
		dataset = np.array(dataset, dtype='int32')

		# Now move the dataset onto GPU
		self.loaded_dataset.set_value(dataset)

		# Return the number of batches
		return num_batches



class Word2VecMinibatcher(NoiseSymbolicMinibatcherMixin, Minibatcher):

	NOT_DONE = 0
	DONE = 1

	def __init__(
		self,
		files=[],
		directories=[],
		skip=[],
		unigram_dictionary=None,
		noise_ratio=15,
		kernel=[1,2,3,4,5,5,4,3,2,1],
		t = 1.0e-5,
		batch_size = 1000,
		verbose=True,
		num_example_generators=3
	):

		# Need to register noise ratio before call to super, because
		# super will call an overridden method that expects noise ratio
		# to be set
		self.noise_ratio = noise_ratio

		super(Word2VecMinibatcher, self).__init__(
			files,
			directories,
			skip,
			batch_size,
			num_example_generators,
			verbose
		)

		# Register parameters not already registered by
		# `super().__init__().`
		self.kernel = kernel
		self.t = t

		# Load the unigram_dictionary if any, or construct a new empty one.
		if unigram_dictionary is not None:
			self.unigram_dictionary = unigram_dictionary
		else:
			self.unigram_dictionary = UnigramDictionary()

		# Validate the kernel.  It should reflect the relative
		# frequencies of choosing tokens from a window of +/- K tokens
		# relative to a query token.  So it must have an even number of
		# entries
		if not len(self.kernel) % 2 == 0:
			raise ValueError(
				'kernel should reflect the relative frequencies of '
				'selecting a context token within +/- K of the query '
				'token, and so should have an equal number of entries '
				'defining frequencies to the left and right of the query '
				'token, and so should have an even number of entries.'
			)


	def get_vocab_size(self):
		'''
		Get the size of the vocabulary.  Only makes sense to call this
		after Minibatcher.prepare() has been called, or if an
		existing (pre-filled) UnigramDictionary was loaded, since otherwise
		it would just return 0.

		(no INPUTS)

		OUTPUTS
		* [int]: size of vocabulary (including `UNK`).

		'''
		# Delegate to the underlying UnigramDictionary
		return len(self.unigram_dictionary)


	def load(self, directory):
		'''
		Loads the unigram_dictionary from files stored in the supplied
		directory.

		INPUTS
		* directory [str]: Path to a directory in which unigram_dictionary
			files are stored.  Unigram dictionary will look for default
			filenames within that directory.

		OUTPUTS
		* [None]
		'''
		# Delegate to the underlying UnigramDictionary
		self.unigram_dictionary.load(directory)


	def save(self, directory):
		'''
		Save the unigram_dictionary to the given directory.
		'''
		# Delegate to the underlying UnigramDictionary
		self.unigram_dictionary.save(directory)


	def preparation(self, savedir, min_frequency=None):
		# Read through the corpus, building the UnigramDictionary
		for filename in self.generate_filenames():
			for tokens in self.parse(filename):
				self.unigram_dictionary.update(tokens)

		# Prune the dictionary, if requested to do so.
		if min_frequency is not None:
			self.unigram_dictionary.prune(min_frequency)


	def prune(self, min_frequency=5):
		'''
		Exposes the prune function for the underlying UnigramDictionary
		'''
		self.unigram_dictionary.prune(min_frequency)


	def batch_examples(self, example_iterator):

		signal_batch, noise_batch = self.init_batch()

		# i keeps track of position in the signal batch
		i = -1
		for signal_example, noise_examples in example_iterator:

			# Increment position within the batch
			i += 1

			# Add the signal example
			signal_batch[i, :] = signal_example

			# Figure out the position within the noise batch
			j = i*self.noise_ratio

			# block-assign the noise samples to the noise batch array
			noise_batch[j:j+self.noise_ratio, :] = noise_examples

			# Once we've finished assembling a minibatch, enqueue it
			# and start assembling a new minibatch
			if i == self.batch_size - 1:
				yield (signal_batch, noise_batch)
				signal_batch, noise_batch = self.init_batch()
				i = -1

		# Normally we'll have a partially filled minibatch after processing
		# the corpus.  The elements in the batch that weren't overwritten
		# contain UNK tokens, which act as padding.  Yield the partial
		# minibatch.
		if i >= 0:
			yield (signal_batch, noise_batch)



	def init_batch(self):
		# Initialize np.array's to store the minibatch data.  We know
		# how big the batch is ahead of time.  Initialize by filling
		# the arrays with UNK tokens.  Doing this means that, at the end
		# of the corpus, when we don't necessarily have a full minibatch,
		# the final minibatch is padded with UNK tokens in order to be
		# of the desired shape.  This has no effect on training, because
		# we don't care about the embedding of the UNK token
		signal_batch = np.full(
			(self.batch_size, 2), UNK, dtype='int32'
		)
		noise_batch = np.full(
			(self.batch_size * self.noise_ratio, 2), UNK, dtype='int32'
		)
		return signal_batch, noise_batch


	def build_examples(self, parsed):

		chooser = TokenChooser(K=len(self.kernel)/2, kernel=self.kernel)

		for tokens in parsed:

			# Isolated tokens (e.g. one-word sentences) have no context
			# and can't be used for training.
			if len(tokens) < 2:
				continue

			token_ids = self.unigram_dictionary.get_ids(tokens)

			# We'll now generate generate signal examples and noise
			# examples for training
			for query_token_pos, query_token_id in enumerate(token_ids):

				# Possibly discard the token
				if self.do_discard(query_token_id):
					continue

				# Sample a token from the context
				context_token_pos = chooser.choose_token(
					query_token_pos, len(token_ids)
				)
				context_token_id = token_ids[context_token_pos]
				signal_example = [query_token_id, context_token_id]

				# Sample tokens from the noise
				noise_context_ids = self.unigram_dictionary.sample(
					(self.noise_ratio,))

				# block-assign the noise samples to the noise batch array
				noise_examples = [
					[query_token_id, noise_context_id]
					for noise_context_id in noise_context_ids
				]

				# Yield the example
				yield (signal_example, noise_examples)


	def do_discard(self, token_id):
		'''
		This function helps with downsampling of very common words.
		Returns true when the token should be discarded as a query word
		'''
		probability = self.unigram_dictionary.get_probability(token_id)
		discard_probability = 1 - np.sqrt(self.t/probability)
		do_discard = np.random.uniform() < discard_probability

		#if do_discard:
		#	print 'discarding', self.unigram_dictionary.get_token(token_id)

		return do_discard



class TokenChooser(object):

	'''
	This choses which context token should be taken given a window
	of +/- K around a query token
	'''

	def __init__(self, K, kernel):
		if not len(kernel) == 2*K:
			raise ValueError(
				'`kernel` must have 2*K entries, one for '
				'each of the elements within the windows of +/- K tokens.'
			)

		self.K = K
		self.kernel = kernel
		self.samplers = {}
		self.indices = range(-K, 0) + range(1, K+1)


	def choose_token(self, idx, length):
		'''
		Randomly choosea  token according to the kernel supplied
		in the constructor.  Note that when sampling the context near
		the beginning of a sentence, the left part of the context window
		will be truncated.  Similarly, sampling context near the end of
		a sentence leads to truncation of the right part of the context
		window.  Short sentences lead to truncation on both sides.

		To ensure that samples are returned within the possibly truncated
		window, two values define the actual extent of the context to be
		sampled:

		`idx`: index of the query word within the context.  E.g. if the
			valid context is constrained to a sentence, and the query word
			is the 3rd token in the sentence, idx should be 2 (because
			of 0-based indexing)

		`length`: length of the the context, E.g. If context is
			constrained to a sentence, and sentence is 7 tokens long,
			length should be 7.
		'''

		# If the token is near the edges of the context, then the
		# sampling kernel will be truncated (we can't sample before the
		# firs word in the sentence, or after the last word).
		# Determine the slice indices that define the truncated kernel.
		negative_idx = length - idx
		start = max(0, self.K - idx)
		stop = min(2*self.K, self.K + negative_idx - 1)

		# We make a separate multinomial sampler for each different
		# truncation of the kernel, because they each define a different
		# set of sampling probabilityes.  If we don't have a sampler for
		# this particular kernel shape, make one.
		if not (start, stop) in self.samplers:

			trunc_probabilities = self.kernel[start:stop]
			self.samplers[start,stop] = (
				Categorical(trunc_probabilities)
			)

		# Sample from the multinomial sampler for the context of this shape
		outcome_idx = self.samplers[start,stop].sample()

		# Map this into the +/- indexing relative to the query word
		relative_idx = self.indices[outcome_idx + start]

		# And then map this into absolute indexing
		result_idx = relative_idx + idx

		return result_idx
